# -*- coding: utf-8 -*-
"""
@author: Oleg Ovchinnikov
"""
from __future__ import division, print_function, absolute_import


def apply_select_channel(file_in_h5, img_num, channel_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_%02i" % (img_num, channel_num)
    image_path = "%s/Raw_Data" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img2)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]

    image_path = "/Frame_%04i/Channel_Current/Filter_Step_0000" % img_num

    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path] = img2
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "Original"
    h5_new_attrs["Number_Of_Variables"] = 0

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    image_path = "/Frame_%04i/Channel_Current" % img_num

    path_main = main_h5_handle.get(image_path)
    path_attrs = path_main

    channel_name = "Channel_%02i" % channel_num
    path_main = main_h5_handle.get(image_path)
    path_attrs = path_main.attrs
    path_attrs["Origin"] = current_ref
    path_attrs["Origin_Region"] = current_reg
    path_attrs["Origin_Name"] = channel_name

    main_h5_handle.close()
    return 1


def apply_wiener_filter(file_in_h5, img_num, filter_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img2)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2), 0]

    # test shape doe wiener filter
    s = 1.1  # constant for now change latter

    size = img.shape
    h = np.empty(size)
    width = size[0]
    height = size[1]
    for x in range(0, width):
        for y in range(0, height):
            h[x, y] = (s / 2 * 3.141592) ** (-s * np.sqrt(x ** 2 + y ** 2))
    h = h / np.sum(np.sum(h))
    H = np.fft.fft2(h)
    K = np.linspace(.001, 1, 100)
    G = np.fft.fft2(img)
    errorV = np.empty([1, 100])
    errorV = errorV[0]
    for k1 in range(0, 100):
        W = np.conj(H) / (np.abs(H) ** 2 + K[k1])
        F = W * G
        R = np.abs(np.fft.ifft2(F))
        errorV[k1] = np.std(np.reshape(R, [1, width * height]))

    # use best option 
    minv = np.min(errorV)
    minl = np.where(errorV == minv)
    minl = minl[0]
    minl = minl[0]
    W = np.conj(H) / (np.abs(H) ** 2 + K[minl])
    F = W * G
    img = np.fft.ifft2(F)

    img = img.reshape(len(img2), 1)
    img = abs(img)
    img = np.real(img)

    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 2):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path] = img
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "Wiener"
    h5_new_attrs["Number_Of_Variables"] = 0

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return 1


def apply_gaussian_corr_filter(file_in_h5, img_num, filter_num, gauss_width, gauss_box_width):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img2)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2), 0]

    gauss_cell = [1, 0, 0, gauss_width, gauss_width, 0]

    s1 = len(img[:, 0])
    s2 = len(img[0, :])
    new_deconv = np.zeros([s1, s2])

    for k1 in range(0, s1):
        for k2 in range(0, s2):
            x_min = max([-gauss_box_width, -k1])
            x_max = min([gauss_box_width, (s1 - k1 - 1)])
            y_min = max([-gauss_box_width, -k2])
            y_max = min([gauss_box_width, (s2 - k2 - 1)])
            [xx, yy] = np.meshgrid(range(y_min, y_max + 1), range(x_min, x_max + 1))
            gaus = fun_2d_gaussian(xx, yy, gauss_cell)
            temp = np.corrcoef(gaus.reshape([1, gaus.size]),
                               img[k1 + x_min:k1 + x_max + 1,
                               k2 + y_min:k2 + y_max + 1].reshape([1, gaus.size]))
            new_deconv[k1, k2] = temp[0, 1]

    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 2):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path] = new_deconv
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "2D Correlation"
    h5_new_attrs["Number_Of_Variables"] = 2
    h5_new_attrs["Variable_1_Name"] = "Filter width"
    h5_new_attrs["Variable_1_Value"] = gauss_box_width
    h5_new_attrs["Variable_2_Name"] = "Gassian width"
    h5_new_attrs["Variable_2_Value"] = gauss_width

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return 1


def fun_2d_gaussian(x, y, parm):
    import numpy as np

    amp = np.double(parm[0])

    x_cent = np.double(parm[1])
    y_cent = np.double(parm[2])

    x_wid = np.double(parm[3])
    y_wid = np.double(parm[4])

    ang = np.double(parm[5])

    a = ((np.cos(ang) ** 2) / (2 * x_wid ** 2)) + ((np.sin(ang) ** 2) / (2 * y_wid ** 2))
    b = -((np.sin(2 * ang)) / (4 * x_wid ** 2)) + ((np.sin(2 * ang)) / (4 * y_wid ** 2))
    c = ((np.sin(ang) ** 2) / (2 * x_wid ** 2)) + ((np.cos(ang) ** 2) / (2 * y_wid ** 2))

    gaussian = amp * (
        np.exp(-((a * (x - x_cent) ** 2) + (2 * b * (x - x_cent) * (y - y_cent)) + (c * (y - y_cent) ** 2))))

    return gaussian


def apply_invert_filter(file_in_h5, img_num, filter_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img)]

    m_img = img.mean()
    img = img - m_img
    img = -img
    img = img + m_img

    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 2):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    temp = 0
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path] = img
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "invert z contrast"
    h5_new_attrs["Number_Of_Variables"] = 0

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return 1


def apply_find(file_path_h5, file_name_h5, file_path_png, file_name_png, filter_width, img_num, filter_num):
    import numpy as np
    import h5py as h5

    image_path = "/Frame_%04i/Filtered_Data/Stack_0000" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    s = 1.1  # constant for now change latter

    size = img.shape
    s1 = size[0]
    s2 = size[1]

    mat_large = np.empty([s1 + 2 * filter_width + 1, s2 + 2 * filter_width + 1])
    mat_large[:, :] = np.inf

    for k1 in range(-filter_width, filter_width + 1):
        for k2 in range(-filter_width, filter_width + 1):
            mat_large[filter_width - k1:-(filter_width + k1) - 1,
                      filter_width - k2:-(filter_width + k2) - 1] = np.minimum(mat_large[filter_width - k1:
                                                                                         -filter_width - k1 - 1,
                                                                                         filter_width - k2:
                                                                                         -filter_width - k2 - 1],
                                                                                h5_image)

    deconv_mat_temp = mat_large[filter_width:len(mat_larg[1, :]) - filter_width,
                      filter_width:len(mat_larg[:, 1]) - filter_width]
    filtered_image = h5_image - deconv_mat_temp

    return 1


def apply_binarization_filter(file_in_h5, img_num, filter_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]

    i_max = img.max()
    i_min = img.min()
    i_diff = i_max - i_min

    max_r = 49
    filter_img = np.zeros([max_r + 1, len(img)])
    time_out = np.zeros([max_r + 1, 1])
    time_out_i = np.zeros([max_r + 1, 1])

    for x in range(0, max_r + 1):
        temp = np.zeros([len(img), 1])
        r = x / max_r
        temp[img > (i_min + (i_diff * r))] = 1
        filter_img[x, :] = temp[:, 0]
        time_out[x] = r
        time_out_i[x] = x + 1

    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 2):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path_f = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path_f] = img
    h5_image_new = main_h5_handle.get(image_path_f)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "Threshold_Reference_Image"
    h5_new_attrs["Number_Of_Variables"] = 0

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    image_path_sv = "%s/Spectroscopic_Values" % image_path
    main_h5_handle[image_path_sv] = time_out
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_sv_ref = h5_image_new.ref
    new_sv_reg = h5_image_new.regionref[0:len(time_out)]

    image_path_si = "%s/Spectroscopic_Indices" % image_path
    main_h5_handle[image_path_si] = time_out_i
    h5_image_new = main_h5_handle.get(image_path_si)
    new_si_ref = h5_image_new.ref
    new_si_reg = h5_image_new.regionref[0:len(time_out)]

    image_path_b = "%s/Binary_Matrix" % image_path
    main_h5_handle[image_path_b] = filter_img
    h5_image_new = main_h5_handle.get(image_path_b)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "Threshold_Matrix"
    h5_new_attrs["Number_Of_Variables"] = 0

    h5_new_attrs["Spectroscopic_Indices"] = new_si_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = new_si_reg
    h5_new_attrs["Spectroscopic_Values"] = new_sv_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = new_sv_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return 1


def apply_binarization_filter_select(file_in_h5, img_num, filter_num, threshold):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(img)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]

    i_max = img.max()
    i_min = img.min()
    i_diff = i_max - i_min

    max_r = 49.0
    filter_img = np.zeros([1, len(img)])
    temp = np.zeros([len(img), 1])
    r = threshold / max_r
    temp[img > (i_min + (i_diff * r))] = 1
    filter_img = temp[:, 0]

    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 2):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)

    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp = 1

    image_path = "%s/Filtered_Image" % image_path
    main_h5_handle[image_path] = filter_img
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "Threshold"
    h5_new_attrs["Number_Of_Variables"] = 1
    h5_new_attrs["Variable_1_Name"] = "Threshold"
    h5_new_attrs["Variable_1_Value"] = r

    h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
    h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
    h5_new_attrs["Position_Indices"] = pos_i_ref
    h5_new_attrs["Position_Indices_Region"] = pos_i_reg
    h5_new_attrs["Position_Values"] = pos_v_ref
    h5_new_attrs["Position_Values_Region"] = pos_v_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return


def cluster_into_atomic_columns(file_in_h5, img_num, filter_num, dist_val):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for ifilt in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, ifilt)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")
    current_ref_m = h5_image.ref
    current_reg_m = h5_image.regionref[0:len(img2)]

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    try:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2), 0]
    except:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2)]

    centers = cluster_2d_oleg_return_geo_center(img, dist_val)

    image_path_org = "/Frame_%04i/Channel_Current" % img_num
    image_path_new = "/Frame_%04i/Channel_Finished" % img_num

    try:
        main_h5_handle.__delitem__(image_path_new)
    except:
        temp = 1

    old_folder = main_h5_handle.get(image_path_org)
    main_h5_handle.create_group(image_path_new)
    new_folder = main_h5_handle.get(image_path_new)
    new_folder_attrs = new_folder.attrs

    parrent_ref = old_folder.attrs.get("Origin")
    parrent_reg = old_folder.attrs.get("Origin_Region")
    parrent_name = old_folder.attrs.get("Origin_Name")
    new_folder_attrs["Origin"] = parrent_ref
    new_folder_attrs["Origin_Region"] = parrent_reg
    new_folder_attrs["Origin_Name"] = parrent_name

    for ifilt in range(0, filter_num + 1):
        image_path_org = "%s/Filter_Step_%04i" % (image_path_org, ifilt)
        image_path_new = "%s/Filter_Step_%04i" % (image_path_new, ifilt)
        image_path_org_temp = "%s/Filtered_Image" % image_path_org
        image_path_new_temp = "%s/Filtered_Image" % image_path_new

        h5_image_old = main_h5_handle.get(image_path_org_temp)
        img_old = np.empty(h5_image_old.shape, dtype=h5_image_old.dtype)
        h5_image_old.read_direct(img_old)

        sec_i_ref = h5_image_old.attrs.get("Spectroscopic_Indices")
        sec_i_reg = h5_image_old.attrs.get("Spectroscopic_Indices_Region")
        sec_v_ref = h5_image_old.attrs.get("Spectroscopic_Values")
        sec_v_reg = h5_image_old.attrs.get("Spectroscopic_Values_Region")
        pos_i_ref = h5_image_old.attrs.get("Position_Indices")
        pos_i_reg = h5_image_old.attrs.get("Position_Indices_Region")
        pos_v_ref = h5_image_old.attrs.get("Position_Values")
        pos_v_rexg = h5_image_old.attrs.get("Position_Values_Region")
        filter_name = h5_image_old.attrs.get("Filter_Name")
        number_var = h5_image_old.attrs.get("Number_Of_Variables")

        main_h5_handle[image_path_new_temp] = img_old
        h5_image_new = main_h5_handle.get(image_path)
        h5_new_attrs = h5_image_new.attrs
        h5_new_attrs["Filter_Name"] = filter_name
        h5_new_attrs["Number_Of_Variables"] = number_var

        for ivar in range(1, number_var + 1):
            var_name = h5_image_old.attrs.get("Variable_%01i_Name" % ivar)
            var_value = h5_image_old.attrs.get("Variable_%01i_Value" % ivar)
            h5_new_attrs["Variable_%01i_Name" % ivar] = var_name
            h5_new_attrs["Variable_%01i_Value" % ivar] = var_value

        h5_new_attrs["Spectroscopic_Indices"] = sec_i_ref
        h5_new_attrs["Spectroscopic_Indices_Region"] = sec_i_reg
        h5_new_attrs["Spectroscopic_Values"] = sec_v_ref
        h5_new_attrs["Spectroscopic_Values_Region"] = sec_v_reg
        h5_new_attrs["Position_Indices"] = pos_i_ref
        h5_new_attrs["Position_Indices_Region"] = pos_i_reg
        h5_new_attrs["Position_Values"] = pos_v_ref
        h5_new_attrs["Position_Values_Region"] = pos_v_reg
        h5_new_attrs["Parent"] = parrent_ref
        h5_new_attrs["Parent_Region"] = parrent_reg

        parrent_ref = h5_image_new.ref
        parrent_reg = h5_image_new.regionref[0:len(img_old)]

    image_path = "%s/Lattice/Positions" % image_path_new
    main_h5_handle[image_path] = centers
    h5_image_new = main_h5_handle.get(image_path)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Method"] = "Density_Clustering"
    h5_new_attrs["Number_Of_Variables"] = 1
    h5_new_attrs["Variable_1_Name"] = "Connection_Distance"
    h5_new_attrs["Variable_1_Value"] = dist_val
    h5_new_attrs["Parent"] = current_ref_m
    h5_new_attrs["Parent_Region"] = current_reg_m

    main_h5_handle.close()


def cluster_2d_oleg(mat_in, dist_val):
    import numpy as np

    tt = [0, 0]
    clusters = []
    to_cluster = np.argwhere(mat_in)
    to_cluster = to_cluster.tolist()

    while len(to_cluster) > 0:
        clust = []
        final_clust = []
        clust.append(to_cluster[0])
        to_cluster.remove(to_cluster[0])
        while (len(clust) > 0) & (len(to_cluster) > 0):
            tt[0] = 5000.0 * dist_val
            tt[1] = len(to_cluster)
            t1 = min(tt)
            t1 = int(t1)
            to_cluster_t = to_cluster[0:t1]
            dem_diff = abs(to_cluster_t - np.tile(clust[0], [t1, 1]))
            diff_vec = np.argwhere(sum(dem_diff, 1) <= dist_val)
            for px in range(len(diff_vec), 0, -1):
                pt = diff_vec[px - 1]
                clust.append(to_cluster[pt])
                to_cluster.remove(to_cluster[pt])
            final_clust.append(clust[0])
            clust.remove(clust[0])

        if len(clust) > 0:
            while len(clust) < 0:
                final_clust.append(clust[0])
                clust.remove(clust[0])

        clusters.append(final_clust)

    return clusters


def cluster_2d_oleg_return_geo_center(mat_in, dist_val):
    import numpy as np

    tt = [0, 0]
    clusters = []
    to_cluster = np.argwhere(mat_in)
    to_cluster = to_cluster.tolist()

    while len(to_cluster) > 0:
        clust = []
        final_clust = []
        clust.append(to_cluster[0])
        to_cluster.remove(to_cluster[0])
        while (len(clust) > 0) & (len(to_cluster) > 0):
            tt[0] = 5000.0 * dist_val
            tt[1] = len(to_cluster)
            t1 = min(tt)
            t1 = int(t1)
            to_cluster_t = to_cluster[0:t1]
            dem_diff = abs(to_cluster_t - np.tile(clust[0], [t1, 1]))
            diff_vec = np.argwhere(np.sum(dem_diff, axis=1) <= dist_val)
            for px in range(len(diff_vec), 0, -1):
                pt = diff_vec[px - 1]
                clust.append(to_cluster[pt])
                to_cluster.remove(to_cluster[pt])
            final_clust.append(clust[0])
            clust.remove(clust[0])

        if len(clust) > 0:
            while len(clust) < 0:
                final_clust.append(clust[0])
                clust.remove(clust[0])

        if len(final_clust) > 2:
            clusters.append(np.mean(final_clust, 0))

    clusters = np.array(clusters)
    clusters = clusters.tolist()

    return clusters


def return_img(file_in_h5, img_num, filter_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Current" % img_num
    for x in range(0, filter_num + 1):
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    sec_v_ref = h5_image.attrs.get("Spectroscopic_Values")
    sec_v_reg = h5_image.attrs.get("Spectroscopic_Values_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")
    pos_v_ref = h5_image.attrs.get("Position_Values")
    pos_v_reg = h5_image.attrs.get("Position_Values_Region")

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    try:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2), 0]
    except:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2)]
    main_h5_handle.close()

    return img


def return_pos(file_in_h5, img_num):
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Finished" % img_num
    type_ref = type(main_h5_handle.get(image_path))
    temp = 1
    x = -1
    while temp:
        x = x + 1
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
        image_temp = "%s/Lattice" % image_path
        temp_loc = main_h5_handle.get(image_temp)
        if type(temp_loc) == type_ref:
            temp = 0

    image_path = image_temp
    image_path = "%s/Positions" % image_path

    h5_image = main_h5_handle.get(image_path)
    pos = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(pos)

    return pos


def run_PCA_atoms(file_in_h5, img_num, box_width):
    import numpy as np
    import h5py as h5
    from scipy import linalg

    main_h5_handle = h5.File(file_in_h5, 'r+')
    image_path = "/Frame_%04i/Channel_Finished" % img_num
    type_ref = type(main_h5_handle.get(image_path))
    temp = 1
    x = -1
    while temp:
        x = x + 1
        image_path = "%s/Filter_Step_%04i" % (image_path, x)
        image_temp = "%s/Lattice" % image_path
        temp_loc = main_h5_handle.get(image_temp)
        if type(temp_loc) == type_ref:
            temp = 0

    pos_path = image_temp
    image_path = "%s/Positions" % pos_path

    h5_image = main_h5_handle.get(image_path)
    pos = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(pos)
    current_ref = h5_image.ref
    current_reg = h5_image.regionref[0:len(pos), 0:len(pos[0])]

    image_path = "/Frame_%04i/Channel_Current" % img_num
    image_path = "%s/Filter_Step_%04i" % (image_path, 0)
    image_path = "%s/Filtered_Image" % image_path

    h5_image = main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)

    sec_i_ref = h5_image.attrs.get("Spectroscopic_Indices")
    sec_i_reg = h5_image.attrs.get("Spectroscopic_Indices_Region")
    pos_i_ref = h5_image.attrs.get("Position_Indices")
    pos_i_reg = h5_image.attrs.get("Position_Indices_Region")

    spec_ind = main_h5_handle[sec_i_ref]
    spec_ind = spec_ind[sec_i_reg]

    posi_ind = main_h5_handle[pos_i_ref]
    posi_ind = posi_ind[pos_i_reg]
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    try:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2), 0]
    except:
        img[posi_ind[:, 0] - 1, posi_ind[:, 1] - 1] = img2[0:len(img2)]

    sel_vec = np.zeros([len(pos[:, 0]), 1])
    kk = 0
    new_pos = []
    img_vectors = []
    for k1 in range(0, len(pos[:, 0])):
        if (box_width <= pos[k1, 0].round() <= len(img[:, 0]) - box_width and
                        box_width <= pos[k1, 1].round() <= len(img[0, :]) - box_width):
            sel_vec[k1] = 1
            new_pos.append(pos[k1, :])
            vector = img[pos[k1, 0] - box_width:pos[k1, 0] + box_width, pos[k1, 1] - box_width:pos[k1, 1] + box_width]
            img_vectors.append(vector.reshape([(box_width * 2) ** 2]))

    new_pos = array(new_pos)
    img_vectors = array(img_vectors)

    U, S, V = linalg.svd(img_vectors)
    V = V[0:len(S)]

    temp = 1
    x = -1

    while temp:
        x = x + 1
        image_temp = "%s/Analysis_%04i" % (pos_path, x)
        temp_loc = main_h5_handle.get(image_temp)
        if type(temp_loc) != type_ref:
            temp = 0

    image_path = image_temp

    [xx, yy] = np.meshgrid(range(1, len(U) + 1), range(1, len(U[0]) + 1))
    xx = np.reshape(xx, [1, len(U) * len(U[0])])
    yy = np.reshape(yy, [1, len(U) * len(U[0])])
    xy = [xx, yy]
    U = np.reshape(U, [1, len(U) * len(U[0])])

    image_path_sv = "%s/Spectroscopic_Values_01" % image_path
    main_h5_handle[image_path_sv] = [1]
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_sv_ref = h5_image_new.ref
    new_sv_reg = h5_image_new.regionref[0]

    image_path_si = "%s/Spectroscopic_Indices_01" % image_path
    main_h5_handle[image_path_si] = [1]
    h5_image_new = main_h5_handle.get(image_path_si)
    new_si_ref = h5_image_new.ref
    new_si_reg = h5_image_new.regionref[0]

    image_path_sv = "%s/Position_Values_01" % image_path
    main_h5_handle[image_path_sv] = xy
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_pv_ref = h5_image_new.ref
    new_pv_reg = h5_image_new.regionref[0:1, 0:len(xy[0])]

    image_path_si = "%s/Position_Indices_01" % image_path
    main_h5_handle[image_path_si] = xy
    h5_image_new = main_h5_handle.get(image_path_si)
    new_pi_ref = h5_image_new.ref
    new_pi_reg = h5_image_new.regionref[0:1, 0:len(xy[0])]

    image_path_b = "%s/Analysis_Data_01_U" % image_path
    main_h5_handle[image_path_b] = U
    h5_image_new = main_h5_handle.get(image_path_b)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "PCA_Atom_Shape"
    h5_new_attrs["Number_Of_Variables"] = 1
    h5_new_attrs["Variable_1_Name"] = "Box Width"
    h5_new_attrs["Variable_1_Value"] = box_width

    h5_new_attrs["Spectroscopic_Indices"] = new_si_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = new_si_reg
    h5_new_attrs["Spectroscopic_Values"] = new_sv_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = new_sv_reg
    h5_new_attrs["Position_Indices"] = new_pi_ref
    h5_new_attrs["Position_Indices_Region"] = new_pi_reg
    h5_new_attrs["Position_Values"] = new_pv_ref
    h5_new_attrs["Position_Values_Region"] = new_pv_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    [xy] = np.meshgrid(range(1, len(S) + 1))

    image_path_sv = "%s/Spectroscopic_Values_02" % image_path
    main_h5_handle[image_path_sv] = [1]
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_sv_ref = h5_image_new.ref
    new_sv_reg = h5_image_new.regionref[0]

    image_path_si = "%s/Spectroscopic_Indices_02" % image_path
    main_h5_handle[image_path_si] = [1]
    h5_image_new = main_h5_handle.get(image_path_si)
    new_si_ref = h5_image_new.ref
    new_si_reg = h5_image_new.regionref[0]

    image_path_sv = "%s/Position_Values_02" % image_path
    main_h5_handle[image_path_sv] = xy
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_pv_ref = h5_image_new.ref
    new_pv_reg = h5_image_new.regionref[0:len(xy)]

    image_path_si = "%s/Position_Indices_02" % image_path
    main_h5_handle[image_path_si] = xy
    h5_image_new = main_h5_handle.get(image_path_si)
    new_pi_ref = h5_image_new.ref
    new_pi_reg = h5_image_new.regionref[0:len(xy)]

    image_path_b = "%s/Analysis_Data_02_S" % image_path
    main_h5_handle[image_path_b] = S
    h5_image_new = main_h5_handle.get(image_path_b)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "PCA_Atom_Shape"
    h5_new_attrs["Number_Of_Variables"] = 1
    h5_new_attrs["Variable_1_Name"] = "Box Width"
    h5_new_attrs["Variable_1_Value"] = box_width

    h5_new_attrs["Spectroscopic_Indices"] = new_si_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = new_si_reg
    h5_new_attrs["Spectroscopic_Values"] = new_sv_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = new_sv_reg
    h5_new_attrs["Position_Indices"] = new_pi_ref
    h5_new_attrs["Position_Indices_Region"] = new_pi_reg
    h5_new_attrs["Position_Values"] = new_pv_ref
    h5_new_attrs["Position_Values_Region"] = new_pv_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    [xx, yy] = np.meshgrid(range(1, len(V) + 1), range(1, len(V[0]) + 1))
    xx = np.reshape(xx, [1, len(V) * len(V[0])])
    yy = np.reshape(yy, [1, len(V) * len(V[0])])
    xy = [xx, yy]
    V = np.reshape(U, [1, len(U) * len(U[0])])

    image_path_sv = "%s/Spectroscopic_Values_03" % image_path
    main_h5_handle[image_path_sv] = [1]
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_sv_ref = h5_image_new.ref
    new_sv_reg = h5_image_new.regionref[0]

    image_path_si = "%s/Spectroscopic_Indices_03" % image_path
    main_h5_handle[image_path_si] = [1]
    h5_image_new = main_h5_handle.get(image_path_si)
    new_si_ref = h5_image_new.ref
    new_si_reg = h5_image_new.regionref[0]

    image_path_sv = "%s/Position_Values_03" % image_path
    main_h5_handle[image_path_sv] = xy
    h5_image_new = main_h5_handle.get(image_path_sv)
    new_pv_ref = h5_image_new.ref
    new_pv_reg = h5_image_new.regionref[0:1, 0:len(xy[0])]

    image_path_si = "%s/Position_Indices_03" % image_path
    main_h5_handle[image_path_si] = xy
    h5_image_new = main_h5_handle.get(image_path_si)
    new_pi_ref = h5_image_new.ref
    new_pi_reg = h5_image_new.regionref[0:1, 0:len(xy[0])]

    image_path_b = "%s/Analysis_Data_03_V" % image_path
    main_h5_handle[image_path_b] = V1
    h5_image_new = main_h5_handle.get(image_path_b)
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"] = "PCA_Atom_Shape"
    h5_new_attrs["Number_Of_Variables"] = 1
    h5_new_attrs["Variable_1_Name"] = "Box Width"
    h5_new_attrs["Variable_1_Value"] = box_width

    h5_new_attrs["Spectroscopic_Indices"] = new_si_ref
    h5_new_attrs["Spectroscopic_Indices_Region"] = new_si_reg
    h5_new_attrs["Spectroscopic_Values"] = new_sv_ref
    h5_new_attrs["Spectroscopic_Values_Region"] = new_sv_reg
    h5_new_attrs["Position_Indices"] = new_pi_ref
    h5_new_attrs["Position_Indices_Region"] = new_pi_reg
    h5_new_attrs["Position_Values"] = new_pv_ref
    h5_new_attrs["Position_Values_Region"] = new_pv_reg
    h5_new_attrs["Parent"] = current_ref
    h5_new_attrs["Parent_Region"] = current_reg

    main_h5_handle.close()

    return



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 07 11:48:53 2017

@author: Suhas Somnath

"""

from __future__ import division, print_function, absolute_import, unicode_literals

import time as tm
import numpy as np
from collections import Iterable
from .process import Process, parallel_compute
from ..io.microdata import MicroDataset, MicroDataGroup
from ..io.hdf_utils import getH5DsetRefs, getAuxData, copyAttributes, link_as_main, linkRefs, check_for_old
from ..io.translators.utils import build_ind_val_dsets
from ..io.io_hdf5 import ioHDF5
from .fft import get_noise_floor, are_compatible_filters, build_composite_freq_filter
# TODO: implement phase compensation
# TODO: correct implementation of num_pix


class SignalFilter(Process):
    def __init__(self, h5_main, frequency_filters=None, noise_threshold=None, write_filtered=True,
                 write_condensed=False, num_pix=1, phase_rad=0,  **kwargs):
        """
        Filters the entire h5 dataset with the given filtering parameters.

        Parameters
        ----------
        h5_main : h5py.Dataset object
            Dataset to process
        frequency_filters : (Optional) single or list of pycroscopy.fft.FrequencyFilter objects
            Frequency (vertical) filters to apply to signal
        noise_threshold : (Optional) float. Default - None
            Noise tolerance to apply to data. Value must be within (0, 1)
        write_filtered : (Optional) bool. Default - True
            Whether or not to write the filtered data to file
        write_condensed : Optional) bool. Default - False
            Whether or not to write the condensed data in frequency space to file. Use this for datasets that are very
            large but sparse in frequency space.
        num_pix : (Optional) uint. Default - 1
            Number of pixels to use for filtering. More pixels means a lower noise floor and the ability to pick up
            weaker signals. Use only if absolutely necessary. This value must be a divisor of the number of pixels in
            the dataset
        phase_rad : (Optional). float
            Degrees by which the output is rotated with respect to the input to compensate for phase lag.
            This feature has NOT yet been implemented.
        kwargs : (Optional). dictionary
            Please see Process class for additional inputs
        """

        super(SignalFilter, self).__init__(h5_main, **kwargs)

        if frequency_filters is None and noise_threshold is None:
            raise ValueError('Need to specify at least some noise thresholding / frequency filter')

        if noise_threshold is not None:
            if noise_threshold >= 1 or noise_threshold <= 0:
                raise ValueError('Noise threshold must be within (0 1)')

        self.composite_filter = 1
        if frequency_filters is not None:
            if not isinstance(frequency_filters, Iterable):
                frequency_filters = [frequency_filters]
            if not are_compatible_filters(frequency_filters):
                raise ValueError('frequency filters must be a single or list of FrequencyFilter objects')
            self.composite_filter = build_composite_freq_filter(frequency_filters)
        else:
            write_condensed = False

        if write_filtered is False and write_condensed is False:
            raise ValueError('You need to write the filtered and/or the condensed dataset to the file')

        num_effective_pix = h5_main.shape[0] * 1.0 / num_pix
        if num_effective_pix % 1 > 0:
            raise ValueError('Number of pixels not divisible by the number of pixels to use for FFT filter')

        self.num_effective_pix = int(num_effective_pix)
        self.phase_rad = phase_rad

        self.noise_threshold = noise_threshold
        self.frequency_filters = frequency_filters

        self.write_filtered = write_filtered
        self.write_condensed = write_condensed

        """
        Remember that the default number of pixels corresponds to only the raw data that can be held in memory
        In the case of signal filtering, the datasets that will occupy space are:
        1. Raw, 2. filtered (real + freq space copies), 3. Condensed (substantially lesser space)
        The actual scaling of memory depends on options:
        """
        scaling_factor = 1 + 2 * self.write_filtered + 0.25 * self.write_condensed
        self._max_pos_per_read = int(self._max_pos_per_read / scaling_factor)

        if self.verbose:
            print('Allowed to read {} pixels per chunk'.format(self._max_pos_per_read))

        self.parms_dict = dict()
        if self.frequency_filters is not None:
            for filter in self.frequency_filters:
                self.parms_dict.update(filter.get_parms())
        if self.noise_threshold is not None:
            self.parms_dict['noise_threshold'] = self.noise_threshold
        self.parms_dict['num_pix'] = self.num_effective_pix

        self.process_name = 'FFT_Filtering'
        self.duplicate_h5_groups = self._check_for_duplicates()

        self.data = None
        self.filtered_data = None
        self.condensed_data = None
        self.noise_floors = None
        self.h5_filtered = None
        self.h5_condensed = None
        self.h5_noise_floors = None

    def _set_memory_and_cores(self, cores=1, mem=1024):
        """
        Checks hardware limitations such as memory, # cpus and sets the recommended datachunk sizes and the
        number of cores to be used by analysis methods.

        Parameters
        ----------
        cores : uint, optional
            Default - 1
            How many cores to use for the computation
        mem : uint, optional
            Default - 1024
            The amount a memory in Mb to use in the computation
        """
        verbose = self.verbose
        self.verbose = False

        super(SignalFilter, self)._set_memory_and_cores(cores, mem)

        self.verbose = verbose

        max_data_chunk = self._max_mem_mb / self._cores

        # Now calculate the number of positions that can be stored in memory in one go.
        # Mem for initial data and final data before writing to file
        mb_per_position = self.h5_main.dtype.itemsize * self.h5_main.shape[1] / 1e6 * 2
        # Mem for fft of data
        mb_per_position += np.complex128.dtype.__sizeof__() * self.h5_main.shape[1] / 1e6

        self._max_pos_per_read = max_data_chunk // mb_per_position

        if self.verbose:
            print('Allowed to read {} pixels per chunk'.format(self._max_pos_per_read))
            print('Allowed to use up to', str(self._cores), 'cores and', str(self._max_mem_mb), 'MB of memory')

    def _create_results_datasets(self):
        """
        Creates all the datasets necessary for holding all parameters + data.
        """

        grp_name = self.h5_main.name.split('/')[-1] + '-' + self.process_name + '_'
        grp_filt = MicroDataGroup(grp_name, self.h5_main.parent.name)

        self.parms_dict.update({'last_pixel': 0, 'algorithm': 'pycroscopy_SignalFilter'})
        grp_filt.attrs = self.parms_dict

        if isinstance(self.composite_filter, np.ndarray):
            ds_comp_filt = MicroDataset('Composite_Filter', np.float32(self.composite_filter))
            grp_filt.addChildren([ds_comp_filt])

        if self.noise_threshold is not None:
            ds_noise_floors = MicroDataset('Noise_Floors',
                                           data=np.zeros(shape=(self.num_effective_pix, 1), dtype=np.float32))
            ds_noise_spec_inds, ds_noise_spec_vals = build_ind_val_dsets([1], is_spectral=True,
                                                                         labels=['arb'], units=[''],
                                                                         verbose=self.verbose)
            ds_noise_spec_inds.name = 'Noise_Spectral_Indices'
            ds_noise_spec_vals.name = 'Noise_Spectral_Values'
            grp_filt.addChildren([ds_noise_floors, ds_noise_spec_inds, ds_noise_spec_vals])

        if self.write_filtered:
            ds_filt_data = MicroDataset('Filtered_Data', data=[], maxshape=self.h5_main.maxshape,
                                        dtype=np.float32, chunking=self.h5_main.chunks, compression='gzip')
            grp_filt.addChildren([ds_filt_data])

        self.hot_inds = None

        h5_pos_inds = getAuxData(self.h5_main, auxDataName=['Position_Indices'])[0]
        h5_pos_vals = getAuxData(self.h5_main, auxDataName=['Position_Values'])[0]

        if self.write_condensed:
            self.hot_inds = np.where(self.composite_filter > 0)[0]
            self.hot_inds = np.uint(self.hot_inds[int(0.5 * len(self.hot_inds)):])  # only need to keep half the data
            ds_spec_inds, ds_spec_vals = build_ind_val_dsets([int(0.5 * len(self.hot_inds))], is_spectral=True,
                                                             labels=['hot_frequencies'], units=[''],
                                                             verbose=self.verbose)
            ds_spec_vals.data = np.atleast_2d(self.hot_inds)  # The data generated above varies linearly. Override.
            ds_cond_data = MicroDataset('Condensed_Data', data=[],
                                        maxshape=(self.num_effective_pix, len(self.hot_inds)),
                                        dtype=np.complex, chunking=(1, len(self.hot_inds)), compression='gzip')
            grp_filt.addChildren([ds_spec_inds, ds_spec_vals, ds_cond_data])
            if self.num_effective_pix > 1:
                # need to make new position datasets by taking every n'th index / value:
                new_pos_vals = np.atleast_2d(h5_pos_vals[slice(0, None, self.num_effective_pix), :])
                ds_pos_inds, ds_pos_vals = build_ind_val_dsets([int(np.unique(h5_pos_inds[:, dim_ind]).size /
                                                                    self.num_effective_pix)
                                                                for dim_ind in range(h5_pos_inds.shape[1])],
                                                               is_spectral=False,
                                                               labels=h5_pos_inds.attrs['labels'],
                                                               units=h5_pos_inds.attrs['units'], verbose=self.verbose)
                h5_pos_vals.data = np.atleast_2d(new_pos_vals)  # The data generated above varies linearly. Override.
                grp_filt.addChildren([ds_pos_inds, ds_pos_vals])

        if self.verbose:
            grp_filt.showTree()
        hdf = ioHDF5(self.h5_main.file)
        h5_filt_refs = hdf.writeData(grp_filt, print_log=self.verbose)

        if isinstance(self.composite_filter, np.ndarray):
            h5_comp_filt = getH5DsetRefs(['Composite_Filter'], h5_filt_refs)[0]

        if self.noise_threshold is not None:
            self.h5_noise_floors = getH5DsetRefs(['Noise_Floors'], h5_filt_refs)[0]
            self.h5_results_grp = self.h5_noise_floors.parent
            link_as_main(self.h5_noise_floors, h5_pos_inds, h5_pos_vals,
                         getH5DsetRefs(['Noise_Spectral_Indices'], h5_filt_refs)[0],
                         getH5DsetRefs(['Noise_Spectral_Values'], h5_filt_refs)[0])

        # Now need to link appropriately:
        if self.write_filtered:
            self.h5_filtered = getH5DsetRefs(['Filtered_Data'], h5_filt_refs)[0]
            self.h5_results_grp = self.h5_filtered.parent
            copyAttributes(self.h5_main, self.h5_filtered, skip_refs=False)
            if isinstance(self.composite_filter, np.ndarray):
                linkRefs(self.h5_filtered, [h5_comp_filt])

            """link_as_main(self.h5_filtered, h5_pos_inds, h5_pos_vals,
                         getAuxData(h5_main, auxDataName=['Spectroscopic_Indices'])[0],
                         getAuxData(h5_main, auxDataName=['Spectroscopic_Values'])[0])"""

        if self.write_condensed:
            self.h5_condensed = getH5DsetRefs(['Condensed_Data'], h5_filt_refs)[0]
            self.h5_results_grp = self.h5_condensed.parent
            if isinstance(self.composite_filter, np.ndarray):
                linkRefs(self.h5_condensed, [h5_comp_filt])
            if self.noise_threshold is not None:
                linkRefs(self.h5_condensed, [self.h5_noise_floors])

            if self.num_effective_pix > 1:
                h5_pos_inds = getH5DsetRefs(['Position_Indices'], h5_filt_refs)[0]
                h5_pos_vals = getH5DsetRefs(['Position_Values'], h5_filt_refs)[0]

            link_as_main(self.h5_condensed, h5_pos_inds, h5_pos_vals,
                         getH5DsetRefs(['Spectroscopic_Indices'], h5_filt_refs)[0],
                         getH5DsetRefs(['Spectroscopic_Values'], h5_filt_refs)[0])

    def _write_results_chunk(self):
        """
        Writes data chunks back to the file
        """

        pos_slice = slice(self._start_pos, self._end_pos)

        if self.write_condensed:
            self.h5_condensed[pos_slice] = self.condensed_data
        if self.noise_threshold is not None:
            self.h5_noise_floors[pos_slice] = np.atleast_2d(self.noise_floors)
        if self.write_filtered:
            self.h5_filtered[pos_slice] = self.filtered_data

        # Leaving in this provision that will allow restarting of processes
        self.h5_results_grp.attrs['last_pixel'] = self._end_pos

        self.hdf.flush()

        print('Finished processing upto pixel ' + str(self._end_pos) + ' of ' + str(self.h5_main.shape[0]))

        # Now update the start position
        self._start_pos = self._end_pos

    @staticmethod
    def _unit_function():
        return get_noise_floor

    def compute(self, *args, **kwargs):
        """
        Creates placeholders for the results, applies the filers to the data, and writes the output to the file.

        Parameters
        ----------

        Returns
        -------
        h5_results_grp : h5py.Datagroup object
            Datagroup containing all the results

        """
        self._create_results_datasets()

        time_per_pix = 0

        num_pos = self.h5_main.shape[0]

        self._read_data_chunk()
        while self.data is not None:

            t_start = tm.time()

            # get FFT of the entire data chunk
            self.data = np.fft.fftshift(np.fft.fft(self.data, axis=1), axes=1)

            if self.noise_threshold is not None:
                self.noise_floors = parallel_compute(self.data, get_noise_floor, cores=self._cores,
                                                     func_args=[self.noise_threshold])

            if isinstance(self.composite_filter, np.ndarray):
                # multiple fft of data with composite filter
                self.data *= self.composite_filter

            if self.noise_threshold is not None:
                # apply thresholding
                self.data[np.abs(self.data) < np.tile(np.atleast_2d(self.noise_floors), self.data.shape[1])] = 1E-16

            if self.write_condensed:
                # set self.condensed_data here
                self.condensed_data = self.data[:, self.hot_inds]

            if self.write_filtered:
                # take inverse FFT
                self.filtered_data = np.real(np.fft.ifft(np.fft.ifftshift(self.data, axes=1), axis=1))
                if self.phase_rad > 0:
                    # do np.roll on data
                    # self.data = np.roll(self.data, 0, axis=1)
                    pass

            tot_time = np.round(tm.time() - t_start, decimals=2)

            if self.verbose:
                print('Done parallel computing in {} sec or {} sec per pixel'.format(tot_time,
                                                                                     tot_time / self.data.shape[0]))
            if self._start_pos == 0:
                time_per_pix = tot_time / self._end_pos  # in seconds
            else:
                print('Time remaining: {} mins'.format(np.round((num_pos - self._end_pos) * time_per_pix / 60, 2)))

            self._write_results_chunk()
            self._read_data_chunk()

        if self.verbose:
            print('Finished processing the dataset completely')

        return self.h5_results_grp



# -*- coding: utf-8 -*-
"""
Created on Thu May 05 13:29:12 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import
from collections import Iterable
from warnings import warn
import matplotlib.pyplot as plt
import numpy as np
from .fft import get_noise_floor, are_compatible_filters, build_composite_freq_filter
from ..io.io_hdf5 import ioHDF5
from ..io.pycro_data import PycroDataset
from ..io.hdf_utils import getH5DsetRefs, linkRefs, getAuxData, copy_main_attributes
from ..io.microdata import MicroDataGroup, MicroDataset
from ..viz.plot_utils import rainbow_plot, set_tick_font_size
from ..io.translators.utils import build_ind_val_dsets

# TODO: Phase rotation not implemented correctly. Find and use excitation frequency


###############################################################################

def test_filter(resp_wfm, frequency_filters=None, noise_threshold=None,
                excit_wfm=None, central_resp_size=None, show_plots=True, use_rainbow_plots=True,
                verbose=False):
    """
    Filters the provided response with the provided filters.

    Parameters
    ----------
    resp_wfm : array-like, 1D
        Raw response waveform in the time domain
    frequency_filters : (Optional) FrequencyFilter object or list of
        Frequency filters to apply to signal
    noise_threshold : (Optional) float
        Noise threshold to apply to signal
    excit_wfm : (Optional) 1D array-like
        Excitation waveform in the time domain. This waveform is necessary for plotting loops.
    central_resp_size : (Optional) unsigned int
        Number of response sample points from the center of the waveform to show in plots. Useful for SPORC
    show_plots : (Optional) Boolean
        Whether or not to plot FFTs before and after filtering
    use_rainbow_plots : (Optional) Boolean
        Whether or not to plot loops whose color varied as a function of time
    verbose : (Optional) Boolean
        Prints extra debugging information if True.  Default False

    Returns
    -------
    filt_data : 1D numpy float array
        Filtered signal in the time domain
    fig : matplotlib.pyplot.figure object
        handle to the plotted figure if requested, else None
    axes : 1D list of matplotlib.pyplot axis objects
        handles to the axes in the plotted figure if requested, else None
    """

    show_loops = excit_wfm is not None and show_plots

    if frequency_filters is None and noise_threshold is None:
        raise ValueError('Need to specify at least some noise thresholding / frequency filter')

    if noise_threshold is not None:
        if noise_threshold >= 1 or noise_threshold <= 0:
            raise ValueError('Noise threshold must be within (0 1)')

    samp_rate = 1
    composite_filter = 1
    if frequency_filters is not None:
        if not isinstance(frequency_filters, Iterable):
            frequency_filters = [frequency_filters]
        if not are_compatible_filters(frequency_filters):
            raise ValueError('frequency filters must be a single or list of FrequencyFilter objects')
        composite_filter = build_composite_freq_filter(frequency_filters)
        samp_rate = frequency_filters[0].samp_rate

    resp_wfm = np.array(resp_wfm)
    num_pts = resp_wfm.size

    fft_pix_data = np.fft.fftshift(np.fft.fft(resp_wfm))

    if noise_threshold is not None:
        noise_floor = get_noise_floor(fft_pix_data, noise_threshold)[0]
        if verbose:
            print('The noise_floor is', noise_floor)


    if show_plots:
        l_ind = int(0.5 * num_pts)
        if isinstance(composite_filter, np.ndarray):
            r_ind = np.max(np.where(composite_filter > 0)[0])
        else:
            r_ind = num_pts
        if verbose:
            print('The left index is {} and the right index is {}.'.format(l_ind, r_ind))

        w_vec = np.linspace(-0.5 * samp_rate, 0.5 * samp_rate, num_pts) * 1E-3
        if central_resp_size:
            sz = int(0.5 * central_resp_size)
            l_resp_ind = -sz + l_ind
            r_resp_ind = l_ind + sz
        else:
            l_resp_ind = l_ind
            r_resp_ind = num_pts
        if verbose:
            print('The left response index is {} and the right response index is {}.'.format(l_resp_ind, r_resp_ind))

        fig = plt.figure(figsize=(12, 8))
        lhs_colspan = 2
        if show_loops is False:
            lhs_colspan = 4
        else:
            ax_loops = plt.subplot2grid((2, 4), (0, 2), colspan=2, rowspan=2)
        ax_raw = plt.subplot2grid((2, 4), (0, 0), colspan=lhs_colspan)
        ax_filt = plt.subplot2grid((2, 4), (1, 0), colspan=lhs_colspan)
        axes = [ax_raw, ax_filt]
        set_tick_font_size(axes, 14)
    else:
        fig = None
        axes = None

    if show_plots:
        amp = np.abs(fft_pix_data)
        ax_raw.semilogy(w_vec[l_ind:r_ind], amp[l_ind:r_ind], label='Raw')
        if frequency_filters is not None:
            ax_raw.semilogy(w_vec[l_ind:r_ind],
                            (composite_filter[l_ind:r_ind] + np.min(amp)) * (np.max(amp) - np.min(amp)),
                            linewidth=3, color='orange', label='Composite Filter')
        if noise_threshold is not None:
            ax_raw.semilogy(w_vec[l_ind:r_ind], np.ones(r_ind - l_ind) * noise_floor,
                            linewidth=2, color='r', label='Noise Threshold')
        ax_raw.legend(loc='best', fontsize=14)
        ax_raw.set_title('Raw Signal', fontsize=16)
        ax_raw.set_ylabel('Magnitude (a. u.)', fontsize=14)

    fft_pix_data *= composite_filter

    if noise_threshold is not None:
        fft_pix_data[np.abs(fft_pix_data) < noise_floor] = 1E-16  # DON'T use 0 here. ipython kernel dies

    if show_plots:
        ax_filt.semilogy(w_vec[l_ind:r_ind], np.abs(fft_pix_data[l_ind:r_ind]))
        ax_filt.set_title('Filtered Signal', fontsize=16)
        ax_filt.set_xlabel('Frequency(kHz)', fontsize=14)
        ax_filt.set_ylabel('Magnitude (a. u.)', fontsize=14)
        if noise_threshold is not None:
            ax_filt.set_ylim(bottom=noise_floor)  # prevents the noise threshold from messing up plots

    filt_data = np.real(np.fft.ifft(np.fft.ifftshift(fft_pix_data)))

    if verbose:
        print('The shape of the filtered data is {}'.format(filt_data.shape))
        print('The shape of the excitation waveform is {}'.format(excit_wfm.shape))

    if show_loops:
        if use_rainbow_plots:
            rainbow_plot(ax_loops, excit_wfm[l_resp_ind:r_resp_ind], filt_data[l_resp_ind:r_resp_ind] * 1E+3)
        else:
            ax_loops.plot(excit_wfm[l_resp_ind:r_resp_ind], filt_data[l_resp_ind:r_resp_ind] * 1E+3)
        ax_loops.set_title('AI vs AO', fontsize=16)
        ax_loops.set_xlabel('Input Bias (V)', fontsize=14)
        ax_loops.set_ylabel('Deflection (mV)', fontsize=14)
        set_tick_font_size(ax_loops, 14)
        axes.append(ax_loops)
        fig.tight_layout()
    return filt_data, fig, axes


###############################################################################


def decompress_response(f_condensed_mat, num_pts, hot_inds):
    """
    Returns the time domain representation of waveform(s) that are compressed in the frequency space
    
    Parameters
    ----------
    f_condensed_mat : 1D or 2D complex numpy arrays
        Frequency domain signals arranged as [position, frequency]. 
        Only the positive frequncy bins must be in the compressed dataset. 
        The dataset is assumed to have been FFT shifted (such that 0 Hz is at the center).
    num_pts : unsigned int
        Number of points in the time domain signal
    hot_inds : 1D unsigned int numpy array
        Indices of the frequency bins in the compressed data. 
        This index array will be necessary to reverse map the condensed 
        FFT into its original form
        
    Returns
    -------
    time_resp : 2D numpy array
        Time domain response arranged as [position, time]
        
    Notes
    -----
    Memory is given higher priority here, so this function loops over the position
    instead of doing the inverse FFT on the complete data.

    """
    f_condensed_mat = np.atleast_2d(f_condensed_mat)
    hot_inds_mirror = np.flipud(num_pts - hot_inds)
    time_resp = np.zeros(shape=(f_condensed_mat.shape[0], num_pts), dtype=np.float32)
    for pos in range(f_condensed_mat.shape[0]):
        f_complete = np.zeros(shape=num_pts, dtype=np.complex)
        f_complete[hot_inds] = f_condensed_mat[pos, :]
        # Now add the mirror (FFT in negative X axis that was removed)
        f_complete[hot_inds_mirror] = np.flipud(f_condensed_mat[pos, :])
        time_resp[pos, :] = np.real(np.fft.ifft(np.fft.ifftshift(f_complete)))

    return np.squeeze(time_resp)


def reshape_from_lines_to_pixels(h5_main, pts_per_cycle, scan_step_x_m=1):
    """
    Breaks up the provided raw G-mode dataset into lines and pixels (from just lines)

    Parameters
    ----------
    h5_main : h5py.Dataset object
        Reference to the main dataset that contains the raw data that is only broken up by lines
    pts_per_cycle : unsigned int
        Number of points in a single pixel
    scan_step_x_m : float
        Step in meters for pixels

    Returns
    -------
    h5_resh : h5py.Dataset object
        Reference to the main dataset that contains the reshaped data
    """
    if h5_main.shape[1] % pts_per_cycle != 0:
        warn('Error in reshaping the provided dataset to pixels. Check points per pixel')
        raise ValueError

    num_cols = int(h5_main.shape[1] / pts_per_cycle)

    h5_spec_vals = getAuxData(h5_main, auxDataName=['Spectroscopic_Values'])[0]
    h5_pos_vals = getAuxData(h5_main, auxDataName=['Position_Values'])[0]
    single_AO = h5_spec_vals[:, :pts_per_cycle]

    ds_spec_inds, ds_spec_vals = build_ind_val_dsets([single_AO.size], is_spectral=True,
                                                     labels=h5_spec_vals.attrs['labels'],
                                                     units=h5_spec_vals.attrs['units'], verbose=False)
    ds_spec_vals.data = np.atleast_2d(single_AO)  # The data generated above varies linearly. Override.

    ds_pos_inds, ds_pos_vals = build_ind_val_dsets([num_cols, h5_main.shape[0]], is_spectral=False,
                                                   steps=[scan_step_x_m, h5_pos_vals[1, 0]],
                                                   labels=['X', 'Y'], units=['m', 'm'], verbose=False)

    ds_reshaped_data = MicroDataset('Reshaped_Data', data=np.reshape(h5_main.value, (-1, pts_per_cycle)),
                                    compression='gzip', chunking=(10, pts_per_cycle))

    # write this to H5 as some form of filtered data.
    resh_grp = MicroDataGroup(h5_main.name.split('/')[-1] + '-Reshape_', parent=h5_main.parent.name)
    resh_grp.addChildren([ds_reshaped_data, ds_pos_inds, ds_pos_vals, ds_spec_inds, ds_spec_vals])

    hdf = ioHDF5(h5_main.file)
    print('Starting to reshape G-mode line data. Please be patient')
    h5_refs = hdf.writeData(resh_grp)

    h5_resh = getH5DsetRefs(['Reshaped_Data'], h5_refs)[0]
    # Link everything:
    linkRefs(h5_resh,
             getH5DsetRefs(['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values'],
                           h5_refs))

    # Copy the two attributes that are really important but ignored:
    copy_main_attributes(h5_main, h5_resh)

    print('Finished reshaping G-mode line data to rows and columns')

    return PycroDataset(h5_resh)



# -*- coding: utf-8 -*-
"""
Created on Thu Jan 12 15:31:55 2017

@author: Suhas Somnath, Kody Law, Rama K. Vasudevan
"""

from __future__ import division, print_function, absolute_import
import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import sqrtm

from ..io.hdf_utils import getAuxData
from ..viz.plot_utils import set_tick_font_size


def do_bayesian_inference(i_meas, bias, freq, num_x_steps=251, r_extra=110, gam=0.03, e=10.0, sigma=10., sigmaC=1.,
                          num_samples=2E3, show_plots=False, econ=False):
    """
    this function accepts a Voltage vector and current vector
    and returns a Bayesian inferred result for R(V) and capacitance
    Used for solving the situation I = V/R(V) + CdV/dt
    to recover R(V) and C, where C is constant.
    Parameters
    ----------
    i_meas : 1D array or list
        current values, should be in nA
    bias : 1D array or list
        voltage values
    freq : float
        frequency of applied waveform
    num_x_steps : unsigned int (Optional, Default = 251)
        Number of steps in x vector (interpolating V)
    r_extra : float (Optional, default = 220 [Ohms])
        Extra resistance in the RC circuit that will provide correct current and resistance values
    gam : float (Optional, Default = 0.03)
        gamma value for reconstruction
    e : float (Optional, Default = 10.0)
        Ask Kody
    sigma : float (Optional, Default = 10.0)
        Ask Kody
    sigmaC : float (Optional, Default = 1.0)
        Ask Kody
    num_samples : unsigned int (Optional, Default = 1E4)
        Number of samples. 1E+4 is more than sufficient
    show_plots : Boolean (Optional, Default = False)
        Whether or not to show plots
    econ : Boolean (Optional, Default = False)
        Whether or not extra datasets are returned. Turn this on when running on multiple datasets
    Returns
    -------
    results_dict : Dictionary
        Dictionary iterms are
        'x' : 1D float array.  Voltage vector interpolated with num_samples number of points
        'm' : Ask Kody
        'mR' : 1D float array.  Bayesian inference of the resistance. This is the one you want
        'vR' : 2D float array.  varaiance ? of inferred resistance
        'Irec' : 1D array or float.  Reconstructed current without capacitance
        'Sigma' : Ask Kody
        'cValue' : float.  Capacitance value
        'm2R' : Ask Kody
        'SI' : Ask Kody
    Written by Kody J. Law (Matlab) and translated to Python by Rama K. Vasudevan
    """
    num_samples = int(num_samples)
    num_x_steps = int(num_x_steps)
    if num_x_steps % 2 == 0:
        num_x_steps += 1  # Always keep it odd

    # Organize, set up the problem
    t_max = 1. / freq
    t = np.linspace(0, t_max, len(bias))
    dt = t[2] - t[1]
    dv = np.diff(bias) / dt
    dv = np.append(dv, dv[-1])
    max_volts = max(bias)
    # num_x_steps = int(round(2 * round(max_volts / dx, 1) + 1, 0))
    x = np.linspace(-max_volts, max_volts, num_x_steps)
    dx = x[1] - x[0]
    # M = len(x)
    num_volt_points = len(bias)

    # Build A
    A = np.zeros(shape=(num_volt_points, num_x_steps + 1))
    for j in range(num_volt_points):
        ix = int(round(np.floor((bias[j] + max_volts) / dx) + 1))
        ix = min(ix, len(x) - 1)
        ix = max(ix, 1)
        A[j, ix] = bias[j] * (bias[j] - x[ix - 1]) / (x[ix] - x[ix - 1])
        A[j, ix - 1] = bias[j] * (1. - (bias[j] - x[ix - 1]) / (x[ix] - x[ix - 1]))

    A[:, num_x_steps] = dv + r_extra * bias

    # generate simulated observations
    Lapt = (-1. * np.diag((t[:-1]) ** 0, -1) - np.diag(t[:-1] ** 0, 1) + 2. * np.diag(t ** 0, 0)) / dt / dt
    Lapt[0, 0] = 1. / dt / dt
    Lapt[-1, -1] = 1. / dt / dt
    O = (1. / gam ** 2) * (np.eye(num_volt_points))
    # noise_term = np.linalg.lstsq(sqrtm(O),np.random.randn(N,1))[0]
    # y = IV_point
    #  Itrue + noise_term.ravel()

    Lap = (-1. * np.diag((x[:-1]) ** 0, -1) - np.diag(x[:-1] ** 0, 1) + 2. * np.diag(x ** 0, 0)) / dx / dx
    Lap[0, 0] = 1. / dx / dx
    Lap[-1, -1] = 1. / dx / dx

    m0 = 3. * np.ones((num_x_steps, 1))
    m0 = np.append(m0, 0)

    P0 = np.zeros(shape=(num_x_steps + 1, num_x_steps + 1))
    P0[:num_x_steps, :num_x_steps] = 1. / sigma ** 2 * (1. * np.eye(num_x_steps) + np.linalg.matrix_power(Lap, 3))
    P0[num_x_steps, num_x_steps] = 1. / sigmaC ** 2

    Sigma = np.linalg.inv(np.dot(A.T, np.dot(O, A)) + P0)
    m = np.dot(Sigma, (np.dot(A.T, np.dot(O, i_meas)) + np.dot(P0, m0)))

    # Reconstructed current
    Irec = np.dot(A, m)  # This includes the capacitance

    # Draw samples from S
    # SI = (np.matlib.repmat(m[:M], num_samples, 1).T) + np.dot(sqrtm(Sigma[:M, :M]), np.random.randn(M, num_samples))
    SI = np.tile(m[:num_x_steps], (num_samples, 1)).T + np.dot(sqrtm(Sigma[:num_x_steps, :num_x_steps]),
                                                               np.random.randn(num_x_steps, num_samples))
    # approximate mean and covariance of R
    mR = 1. / num_samples * np.sum(1. / SI, 1)
    m2R = 1. / num_samples * np.dot(1. / SI, (1. / SI).T)
    # m2R=1./num_samples*(1./SI)*(1./SI).T
    # vR=m2R-np.dot(mR,mR.T)
    vR = m2R - mR * mR.T
    cValue = m[-1]

    if econ:
        results_dict = {'x': x, 'mR': mR, 'vR': np.diag(vR), 'Irec': Irec, 'cValue': cValue}
    else:
        results_dict = {'x': x, 'm': m, 'mR': mR, 'vR': vR, 'Irec': Irec, 'Sigma': Sigma, 'cValue': cValue, 'm2R': m2R,
                        'SI': SI}

    if show_plots:
        # Do some plotting
        plt.figure(101)
        plt.plot(x, mR, 'b', linewidth=3)
        plt.plot(x, mR + np.sqrt(np.diag(vR)), 'r-', linewidth=3)
        plt.plot(x, mR - np.sqrt(np.diag(vR)), 'r-', linewidth=3)
        plt.xlabel('Voltage (V)')
        plt.ylabel('Resistance (GOhm)')
        plt.title('R(V)')
        plt.legend(('R(V)', 'R(V)+sigma', 'R(V)-sigma'), loc='best')
        # plt.ylim((0,3))
        plt.xlim((-max_volts, max_volts))

        plt.figure(102)
        plt.plot(bias, i_meas)
        plt.plot(x, x / mR)
        plt.xlabel('Voltage')
        plt.ylabel('Current')
        plt.legend(('measured current', 'reconstructed I (no C)'), loc='best')

        plt.figure(103)
        plt.plot(bias, Irec)
        plt.plot(bias, i_meas)
        plt.legend(('I$_{rec}$', 'I$_{true}$'), loc='best')

        plt.figure(104)
        cx = np.arange(0, 2, 0.01)
        dens_cx = 1. / np.sqrt(Sigma[num_x_steps, num_x_steps] * 2 * np.pi) * np.exp(
            -(cx - m[num_x_steps]) ** 2 / 2 / Sigma[num_x_steps, num_x_steps])
        plt.plot(cx, dens_cx)
        plt.ylabel('p(C)')
        plt.xlabel('C')

        print("The value of the capacitance is ", str(round(m[-1] * 1E3, 2)) + "pF")

    return results_dict


def bayesian_inference_on_period(i_meas, excit_wfm, ex_freq, r_extra=110, num_x_steps=500, show_plots=False,
                                 r_max=None, **kwargs):
    """
    Performs Bayesian Inference on a single I-V curve.
    The excitation waveform must be a single period of a sine wave.
    This algorithm splits the curve into the forward and reverse sections, performs inference on each of the sections,
    stitches the results back again, and corrects the resistance which is not handled in the main bayesian function.
    Parameters
    ----------
    i_meas : array-like
        Current corresponding to a single period of sinusoidal excitation bias
    excit_wfm : array-like
        Single period of the sinusoidal excitation waveform
    ex_freq : float
        Frequency of the excitation waveform
    r_extra : float (Optional, default = 220 [Ohms])
        Extra resistance in the RC circuit that will provide correct current and resistance values
    num_x_steps : uint (Optional, default = 500)
        Number of steps for the inferred results. Note: this may be different from what is specified.
    show_plots : Boolean (Optional, Default = False)
        Whether or not to show plots
    r_max : float (Optional, Default = None)
        Maximum limit of the resistance plots.
    kwargs : dict
        Other parameters that will be passed on to the do_bayesian_inference function
    Returns
    -------
    results : dictionary
        Dictionary iterms are
        'cValue' : array-like - 2 elements
            Capacitance on the forward and reverse sections
        'x' : array-like
            Interpolated bias from bayesian inference of length num_x_steps
        'mR' : array-like
            Resistance of sample infered by Bayesian Inference of length num_x_steps
        'vR' : array-like
            Variance of the inferred resistance of length num_x_steps
        'IcorrSine' : array-like
            Measured current with the capacitance correctly subtracted.
        'Irec' : array-like
            Current reconstructed via Bayesian Inference
    """
    roll_val = -0.25
    num_v_steps = excit_wfm.size
    cos_omega_t = np.roll(excit_wfm, int(num_v_steps * roll_val))
    y_val = np.roll(i_meas, int(num_v_steps * roll_val))
    half_x_steps = num_x_steps // 2
    rev_results = do_bayesian_inference(y_val[:int(0.5 * num_v_steps)] * -1,
                                        cos_omega_t[:int(0.5 * num_v_steps)] * -1,
                                        ex_freq, num_x_steps=half_x_steps,
                                        econ=True, show_plots=False, r_extra=r_extra, **kwargs)
    forw_results = do_bayesian_inference(y_val[int(0.5 * num_v_steps):], cos_omega_t[int(0.5 * num_v_steps):],
                                         ex_freq, num_x_steps=half_x_steps,
                                         econ=True, show_plots=False, r_extra=r_extra, **kwargs)

    # putting the split inference together:
    full_results = dict()
    for item in ['cValue']:
        full_results[item] = np.hstack((forw_results[item], rev_results[item]))

    # Capacitance is always doubled - halve it now:
    full_results['cValue'] *= 0.5
    cap_val = np.mean(full_results['cValue'])

    # Compensating the resistance..
    # omega = 2 * np.pi * ex_freq
    """t_max = 1. / ex_freq
    t = np.linspace(0, t_max, len(excit_wfm))
    dt = t[2] - t[1]"""
    # dt = period time / points per period
    dt = 1 / (ex_freq * excit_wfm.size)
    dv = np.diff(excit_wfm) / dt
    dv = np.append(dv, dv[-1])
    i_cap = cap_val * dv
    i_extra = r_extra * 2 * cap_val * excit_wfm
    i_corr_sine = i_meas - i_cap - i_extra
    full_results['IcorrSine'] = i_corr_sine

    # by default Bayesian inference will sort bias in ascending order
    rev_results['x'] *= -1
    rev_results['Irec'] *= -1

    for item in ['x', 'mR', 'vR', 'Irec']:
        full_results[item] = np.hstack((forw_results[item], rev_results[item]))
        # print(item, full_results[item].shape)

    full_results['Irec'] = np.roll(full_results['Irec'], int(num_v_steps * roll_val))

    # Plot to make sure things are indeed correct:
    if show_plots:
        fig, axis = plt.subplots(figsize=(8, 8))
        axis.plot(excit_wfm, i_meas, color='green', label='Meas')
        axis.plot(excit_wfm, i_corr_sine, color='k', label='Sine corr')  # should not be able to see this.
        axis.plot(excit_wfm, i_extra, '--', color='grey', label='I extra')
        axis.plot(excit_wfm, full_results['Irec'], '--', color='orange', label='I rec')
        axis.plot(forw_results['x'], forw_results['x'] / forw_results['mR'], color='blue', label='Bayes corr F')
        axis.plot(rev_results['x'], rev_results['x'] / rev_results['mR'], color='red', label='Bayes corr R')
        axis.set_xlabel('Bias (V)')
        axis.set_ylabel('Current')
        axis.legend()
        axis.axhline(y=0, xmin=np.min(excit_wfm), xmax=np.max(excit_wfm), ls=':')
        fig.tight_layout()

        def _plot_resistance(axis, bias_triang, res_vec, variance_vec, forward=True):
            st_dev = np.sqrt(variance_vec)
            good_pts = np.where(st_dev < 10)[0]
            good_pts = good_pts[np.where(good_pts < forw_results['x'].size)[0]]
            pos_limits = res_vec + st_dev
            neg_limits = res_vec - st_dev
            if forward:
                cols_set = ['blue', 'cyan']
            else:
                cols_set = ['red', 'orange']

            axis.plot(bias_triang[good_pts], res_vec[good_pts], color=cols_set[0], label='R(V)')
            axis.fill_between(bias_triang[good_pts], pos_limits[good_pts], neg_limits[good_pts],
                              alpha=0.25, color=cols_set[1], label='R(V)+-$\sigma$')

        fig, axes = plt.subplots(ncols=2, nrows=1, figsize=(10, 5))
        for axis, res_vec, variance_vec, name, direction in zip(axes.flat,
                                                                [forw_results['mR'], rev_results['mR']],
                                                                [forw_results['vR'], rev_results['vR']],
                                                                ['Forw', 'Rev'], [True, False, ]):
            _plot_resistance(axis, forw_results['x'], res_vec, variance_vec, forward=direction)
            y_lims = axis.get_ylim()
            if r_max is not None:
                axis.set_ylim([0, min(r_max, y_lims[1])])
            else:
                axis.set_ylim([0, y_lims[1]])
            axis.set_title(name)
            axis.set_xlabel('Bias (V)')
            axis.set_ylabel('Resistance')
            axis.legend()
        fig.tight_layout()

    return full_results


def plot_bayesian_spot_from_h5(h5_bayesian_grp, h5_resh, pix_ind, **kwargs):
    """
    Plots the basic Bayesian Inference results for a specific pixel
    Parameters
    ----------
    h5_bayesian_grp : h5py.Datagroup reference
        Group containing the Bayesian Inference results
    h5_resh : h5py.Dataset reference
        Dataset containing the raw / filtered measured current split by pixel
    pix_ind : unsigned int
        Integer index of the desired pixel
    Returns
    -------
    fig : matplotlib.pyplot figure handle
        Handle to figure
    """
    bias_triang = np.squeeze(h5_bayesian_grp['Spectroscopic_Values'][()])
    h5_resistance = h5_bayesian_grp['Resistance']
    h5_r_variance = h5_bayesian_grp['R_variance']
    h5_i_corrected = h5_bayesian_grp['Corrected_Current']

    i_meas = np.squeeze(h5_resh[pix_ind])
    orig_bias = np.squeeze(getAuxData(h5_resh, auxDataName=['Spectroscopic_Values'])[0])
    h5_pos = getAuxData(h5_resh, auxDataName=['Position_Indices'])[0]

    resistance = h5_resistance[pix_ind]
    i_correct = h5_i_corrected[pix_ind]
    r_variance = h5_r_variance[pix_ind]

    return plot_bayesian_results(orig_bias, i_meas, i_correct, bias_triang, resistance, r_variance,
                                 pix_pos=h5_pos[pix_ind], **kwargs)


def plot_bayesian_results(bias_sine, i_meas, i_corrected, bias_triang, resistance, r_variance, i_recon=None,
                          pix_pos=[0, 0], broken_resistance=True, r_max=None, res_scatter=False, **kwargs):
    """
    Plots the basic Bayesian Inference results for a specific pixel
    Parameters
    ----------
    bias_sine : 1D float numpy array
        Original bias vector used for experiment
    i_meas : 1D float numpy array
        Current measured from experiment
    i_corrected : 1D float numpy array
        current with capacitance and R extra compensated
    i_recon : 1D float numpy array
        Reconstructed current
    bias_triang : 1D float numpy array
        Interpolated bias
    resistance : 1D float numpy array
        Inferred resistance
    r_variance : 1D float numpy array
        Variance of the resistance
    pix_pos : list of two numbers
        Pixel row and column positions or values
    broken_resistance : bool, Optional
        Whether or not to break the resistance plots into sections so as to avoid plotting areas with high variance
    r_max : float, Optional
        Maximum value of resistance to plot
    res_scatter : bool, Optional
        Use scatter instead of line plots for resistance
    Returns
    -------
    fig : matplotlib.pyplot figure handle
        Handle to figure
    """

    font_size_1 = 14
    font_size_2 = 16

    half_x_ind = int(0.5 * bias_triang.size)

    ex_amp = np.max(bias_triang)

    colors = [['red', 'orange'], ['blue', 'cyan']]
    syms = [['-', '--', '--'], ['-', ':', ':']]
    names = ['Forward', 'Reverse']

    cos_omega_t = np.roll(bias_sine, int(-0.25 * bias_sine.size))
    orig_half_pt = int(0.5 * bias_sine.size)
    i_correct_rolled = np.roll(i_corrected, int(-0.25 * bias_sine.size))

    st_dev = np.sqrt(r_variance)
    tests = [st_dev < 10, resistance > 0]
    if r_max is not None:
        tests.append(resistance < r_max)
    good_pts = np.ones(resistance.shape, dtype=bool)
    for item in tests:
        good_pts = np.logical_and(good_pts, item)
    good_pts = np.where(good_pts)[0]
    good_forw = good_pts[np.where(good_pts < half_x_ind)[0]]
    good_rev = good_pts[np.where(good_pts >= half_x_ind)[0]]
    pos_limits = resistance + st_dev
    neg_limits = resistance - st_dev

    fig, axes = plt.subplots(ncols=3, figsize=(15, 5))
    # fig.subplots_adjust(wspace=3.5)

    axes[0].set_ylabel('Resistance (G$\Omega$)', fontsize=font_size_2)

    pts_to_plot = [good_forw, good_rev]

    for type_ind, axis, pts_list, cols_set, sym_set, set_name in zip(range(len(names)),
                                                                     axes[:2], pts_to_plot,
                                                                     colors, syms, names):
        axis.set_title('$R(V)$ ' + set_name + ' at Row = ' + str(pix_pos[1]) +
                       ' Col =' + str(pix_pos[0]), fontsize=font_size_2)

        single_plot = not broken_resistance
        if broken_resistance:
            diff = np.diff(pts_list)
            jump_inds = np.argwhere(diff > 4) + 1
            if jump_inds.size < 1:
                single_plot = True

        if not single_plot:
            jump_inds = np.append(np.append(0, jump_inds), pts_list[-1])
            for ind in range(1, jump_inds.size):
                cur_range = pts_list[jump_inds[ind - 1]:jump_inds[ind]]
                if res_scatter:
                    axis.scatter(bias_triang[cur_range], resistance[cur_range],
                                 color=cols_set[0], s=30)
                else:
                    axis.plot(bias_triang[cur_range], resistance[cur_range], cols_set[0],
                              linestyle=sym_set[0], linewidth=3)
                axis.fill_between(bias_triang[cur_range], pos_limits[cur_range], neg_limits[cur_range],
                                  alpha=0.25, color=cols_set[1])
                if ind == 1:
                    axis.legend(['R(V)', 'R(V)+-$\sigma$'], loc='upper center', fontsize=font_size_1)
        else:
            if res_scatter:
                axis.scatter(bias_triang[pts_list], resistance[pts_list],
                             color=cols_set[0], s=30)
            else:
                axis.plot(bias_triang[pts_list], resistance[pts_list], cols_set[0],
                          linestyle=sym_set[0], linewidth=3, label='R(V)')
            axis.fill_between(bias_triang[pts_list], pos_limits[pts_list], neg_limits[pts_list],
                              alpha=0.25, color=cols_set[1], label='R(V)+-$\sigma$')
            axis.legend(loc='upper center', fontsize=font_size_1)
        axis.set_xlabel('Voltage (V)', fontsize=font_size_2)

        axis.set_xlim((-ex_amp, ex_amp))

    # ################### CURRENT PLOT ##########################

    axes[2].plot(bias_sine, i_meas, 'green', linewidth=3, label='I$_{meas}$')
    if i_recon is not None:
        axes[2].plot(bias_sine, i_recon, 'c--', linewidth=3, label='I$_{recon}$')
    axes[2].plot(cos_omega_t[orig_half_pt:], i_correct_rolled[orig_half_pt:],
                 'blue', linewidth=3, label='I$_{Bayes} Forw$')
    axes[2].plot(cos_omega_t[:orig_half_pt], i_correct_rolled[:orig_half_pt],
                 'red', linewidth=3, label='I$_{Bayes} Rev$')

    # axes[2].legend(loc='upper right', bbox_to_anchor=(-.1, 0.30), fontsize=font_size_1)
    axes[2].legend(loc='best', fontsize=font_size_1)
    axes[2].set_xlabel('Voltage(V)', fontsize=font_size_2)
    axes[2].set_title('$I(V)$ at row ' + str(pix_pos[0]) + ', col ' + str(pix_pos[1]),
                      fontsize=font_size_2)

    axes[2].set_ylabel('Current (nA)', fontsize=font_size_2)

    set_tick_font_size(axes, font_size_1)

    fig.tight_layout()

    return fig



"""
Created on Feb 9, 2016

@author: Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import json
import os
import zipfile
from warnings import warn

import numpy as np
from skimage.measure import block_reduce
from skimage.util import crop

from .df_utils.io_image import unnest_parm_dicts, read_dm3
from .translator import Translator
from .utils import generate_dummy_main_parms, make_position_mat, get_spectral_slicing, \
    get_position_slicing, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, getH5GroupRefs, calc_chunks, link_as_main
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class NDataTranslator(Translator):
    """
    Translate Pytchography data from a set of images to an HDF5 file
    """

    def __init__(self, *args, **kwargs):
        super(NDataTranslator, self).__init__(*args, **kwargs)

        self.rebin = False
        self.bin_factor = (1, 1, 1, 1)
        self.hdf = None
        self.binning_func = self.__no_bin
        self.bin_func = None
        self.h5_main = None
        self.root_image_list = list()
        self.crop_method = 'percent'
        self.crop_ammount = None
        self.image_list_tag = None

    def translate(self, h5_path, image_path, bin_factor=None, bin_func=np.mean, start_image=0, scan_size_x=None,
                  scan_size_y=None, crop_ammount=None, crop_method='percent'):
        """
        Basic method that adds Ptychography data to existing hdf5 thisfile
        You must have already done the basic translation with BEodfTranslator

        Parameters
        ----------------
        h5_path : str
            Absolute path to where the HDF5 file should be located
        image_path : str
            Absolute path to folder holding the image files or the path to a specific file
        bin_factor : array_like of uint, optional
            Downsampling factor for each dimension.  Default is None.
        bin_func : callable, optional
            Function which will be called to calculate the return value
            of each block.  Function must implement an axis parameter,
            i.e. numpy.mean.  Ignored if bin_factor is None.  Default is
            numpy.mean.
        start_image : int, optional
            Integer denoting which image in the file path should be considered the starting
            point.  Default is 0, start with the first image on the list.
        scan_size_x : int, optional
            Number of Ronchigrams in the x direction.  Default is None, value will be determined
            from the number of images and `scan_size_y` if it is given.
        scan_size_y : int, optional
            Number of Ronchigrams in the y direction.  Default is None, value will be determined
            from the number of images and `scan_size_x` if it is given.
        crop_ammount : uint or list of uints, optional
            How much should be cropped from the original image.  Can be a single unsigned
            integer or a list of unsigned integers.  A single integer will crop the same
            ammount from all edges.  A list of two integers will crop the x-dimension by
            the first integer and the y-dimension by the second integer.  A list of 4
            integers will crop the image as [left, right, top, bottom].
        crop_method : str, optional
            Which cropping method should be used.  How much of the image is removed is
            determined by the value of `crop_ammount`.
            'percent' - A percentage of the image is removed.
            'absolute' - The specific number of pixel is removed.
        Returns
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset object that contains the flattened images

        """
        # Open the hdf5 file and delete any contents
        try:
            hdf = ioHDF5(h5_path)
            hdf.clear()
        except:
            raise

        self.hdf = hdf
        self.crop_method = crop_method
        self.crop_ammount = crop_ammount

        '''
        Get the list of all files with the .tif extension and
        the number of files in the list
        '''
        file_list = self._parse_file_path(image_path)

        image_parm_list = self._getimageparms(file_list)

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            self.rebin = True
            if isinstance(bin_factor, int):
                self.bin_factor = (1, 1, bin_factor, bin_factor)
            elif len(bin_factor) == 2:
                self.bin_factor = (1, 1) + bin_factor
            else:
                raise ValueError('Input parameter `bin_factor` must be a length 2 array_like or an integer.\n' +
                                 '{} was given.'.format(bin_factor))

            self.binning_func = block_reduce
            self.bin_func = bin_func

        h5_channels = self._setupH5(image_parm_list)

        self._read_data(file_list, h5_channels)

        self.hdf.close()

        return

    def _create_root_image(self, image_path):
        """
        Create the Groups and Datasets for a single root image

        Parameters
        ----------
        image_path : str
            Path to the image file

        Returns
        -------
        None
        """
        image, image_parms = read_dm3(image_path)
        if image.ndim == 3:
            image = np.sum(image, axis=0)

        '''
        Create the Measurement and Channel Groups to hold the
        image Datasets
        '''
        root_grp = MicroDataGroup('/')

        meas_grp = MicroDataGroup('Measurement_')

        chan_grp = MicroDataGroup('Channel_')
        root_grp.addChildren([meas_grp])
        meas_grp.addChildren([chan_grp])

        '''
        Set the Measurement Group attributes
        '''
        meas_grp.attrs.update(image_parms)
        usize, vsize = image.shape
        meas_grp.attrs['image_size_u'] = usize
        meas_grp.attrs['image_size_v'] = vsize
        meas_grp.attrs['translator'] = 'OneView'
        meas_grp.attrs['num_pixels'] = image.size

        ds_rawimage = MicroDataset('Raw_Data', np.reshape(image, (-1, 1)))

        '''
        Build Spectroscopic and Position datasets for the image
        '''
        pos_mat = make_position_mat(image.shape)
        spec_mat = np.array([[0]], dtype=np.uint8)

        ds_spec_inds = MicroDataset('Spectroscopic_Indices', spec_mat)
        ds_spec_vals = MicroDataset('Spectroscopic_Values', spec_mat, dtype=np.float32)
        spec_lab = get_spectral_slicing(['Image'])
        ds_spec_inds.attrs['labels'] = spec_lab
        ds_spec_inds.attrs['units'] = ''
        ds_spec_vals.attrs['labels'] = spec_lab
        ds_spec_vals.attrs['units'] = ''

        ds_pos_inds = MicroDataset('Position_Indices', pos_mat)
        ds_pos_vals = MicroDataset('Position_Values', pos_mat, dtype=np.float32)

        pos_lab = get_position_slicing(['X', 'Y'])
        ds_pos_inds.attrs['labels'] = pos_lab
        ds_pos_inds.attrs['units'] = ['pixel', 'pixel']
        ds_pos_vals.attrs['labels'] = pos_lab
        ds_pos_vals.attrs['units'] = ['pixel', 'pixel']

        chan_grp.addChildren([ds_rawimage, ds_spec_inds, ds_spec_vals,
                              ds_pos_inds, ds_pos_vals])

        '''
        Write the data to file and get the handle for the image dataset
        '''
        image_refs = self.hdf.writeData(root_grp)

        h5_image = getH5DsetRefs(['Raw_Data'], image_refs)[0]

        '''
        Link references to raw
        '''
        aux_ds_names = ['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values']
        link_as_main(h5_image, *getH5DsetRefs(aux_ds_names, image_refs))

        self.root_image_list.append(h5_image)

    def _read_data(self, file_list, h5_channels):
        """
        Iterates over the images in `file_list`, reading each image and downsampling if
        reqeusted, and writes the flattened image to file.  Also builds the Mean_Ronchigram
        and the Spectroscopic_Mean datasets at the same time.

        Parameters
        ----------
        file_list : list of str
            List of all files in `image_path` that will be read
        h5_main : h5py.Dataset
            Dataset which will hold the Ronchigrams
        h5_mean_spec : h5py.Dataset
            Dataset which will hold the Spectroscopic Mean
        h5_ronch : h5py.Dataset
            Dataset which will hold the Mean Ronchigram
        image_path : str
            Absolute file path to the directory which hold the images

        Returns
        -------
        None
        """
        h5_main_list = list()
        '''
        For each file, we must read the data then create the neccessary datasets, add them to the channel, and
        write it all to file
        '''

        '''
        Get zipfile handles for all the ndata1 files that were found in the image_path
        '''

        for ifile, (this_file, this_channel) in enumerate(zip(file_list, h5_channels)):
            _, ext = os.path.splitext(this_file)
            if ext == '.ndata1':
                '''
                Extract the data file from the zip archive and read it into an array
                '''
                this_zip = zipfile.ZipFile(this_file, 'r')
                tmp_path = this_zip.extract('data.npy')
                this_data = np.load(tmp_path)
                os.remove(tmp_path)
            elif ext == '.npy':
                # Read data directly from npy file
                this_data = np.load(this_file)

            '''
            Find the shape of the data, then calculate the final dimensions based on the crop and
            downsampling parameters
            '''
            while this_data.ndim < 4:
                this_data = np.expand_dims(this_data, 0)

            this_data = self.crop_ronc(this_data)
            scan_size_x, scan_size_y, usize, vsize = this_data.shape

            usize = int(round(1.0 * usize / self.bin_factor[-2]))
            vsize = int(round(1.0 * vsize / self.bin_factor[-1]))

            num_images = scan_size_x * scan_size_y
            num_pixels = usize * vsize

            '''
            Write these attributes to the Measurement group
            '''
            new_attrs = {'image_size_u': usize,
                         'image_size_v': vsize,
                         'scan_size_x': scan_size_x,
                         'scan_size_y': scan_size_y}
            this_channel.parent.attrs.update(new_attrs)

            # Get the Position and Spectroscopic Datasets
            ds_spec_ind, ds_spec_vals = build_ind_val_dsets((usize, vsize), is_spectral=True,
                                                            labels=['U', 'V'], units=['pixel', 'pixel'])
            ds_pos_ind, ds_pos_val = build_ind_val_dsets([scan_size_x, scan_size_y], is_spectral=False,
                                                         labels=['X', 'Y'], units=['pixel', 'pixel'])

            ds_chunking = calc_chunks([num_images, num_pixels],
                                      np.float32(0).itemsize,
                                      unit_chunks=(1, num_pixels))

            # Allocate space for Main_Data and Pixel averaged Data
            ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_images, num_pixels),
                                        chunking=ds_chunking, dtype=np.float32, compression='gzip')
            ds_mean_ronch_data = MicroDataset('Mean_Ronchigram',
                                              data=np.zeros(num_pixels, dtype=np.float32),
                                              dtype=np.float32)
            ds_mean_spec_data = MicroDataset('Spectroscopic_Mean',
                                             data=np.zeros(num_images, dtype=np.float32),
                                             dtype=np.float32)

            # Add datasets as children of Measurement_000 data group
            ds_channel = MicroDataGroup(this_channel.name)
            ds_channel.addChildren([ds_main_data, ds_spec_ind, ds_spec_vals, ds_pos_ind,
                                    ds_pos_val, ds_mean_ronch_data, ds_mean_spec_data])

            h5_refs = self.hdf.writeData(ds_channel)
            h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            h5_ronch = getH5DsetRefs(['Mean_Ronchigram'], h5_refs)[0]
            h5_mean_spec = getH5DsetRefs(['Spectroscopic_Mean'], h5_refs)[0]

            aux_ds_names = ['Position_Indices',
                            'Position_Values',
                            'Spectroscopic_Indices',
                            'Spectroscopic_Values']

            link_as_main(h5_main, *getH5DsetRefs(aux_ds_names, h5_refs))

            mean_ronch = np.zeros(h5_ronch.shape, dtype=np.float32)

            this_data = self.binning_func(this_data, self.bin_factor, self.bin_func).reshape(h5_main.shape)

            h5_main[:, :] = this_data

            h5_mean_spec[:] = np.mean(this_data, axis=1)

            h5_ronch[:] = np.mean(this_data, axis=0)

            self.hdf.flush()

            h5_main_list.append(h5_main)

        self.hdf.flush()

    def crop_ronc(self, ronc):
        """
        Crop the input Ronchigram by the specified ammount using the specified method.

        Parameters
        ----------
        ronc : numpy.array
            Input image to be cropped.

        Returns
        -------
        cropped_ronc : numpy.array
            Cropped image
        """

        if self.crop_ammount is None:
            return ronc

        crop_ammount = self.crop_ammount
        crop_method = self.crop_method

        if crop_method == 'percent':
            crop_ammount = np.round(np.atleast_2d(crop_ammount) / 100.0 * ronc.shape)
            crop_ammount = tuple([tuple(row) for row in crop_ammount.astype(np.uint32)])
        elif crop_method == 'absolute':
            if isinstance(crop_ammount, int):
                crop_ammount = ((crop_ammount,), (crop_ammount,))
            elif len(crop_ammount) == 2:
                crop_ammount = ((crop_ammount[0],), (crop_ammount[1],))
            elif len(crop_ammount) == 4:
                crop_ammount = ((crop_ammount[0], crop_ammount[1]), (crop_ammount[2], crop_ammount[3]))
            else:
                raise ValueError('The crop_ammount should be an integer or list of 2 or 4 integers.')
        else:
            raise ValueError('Allowed values of crop_method are percent and absolute.')

        crop_ammount = ((0,), (0,)) + crop_ammount

        cropped_ronc = crop(ronc, crop_ammount)

        if any([dim == 0 for dim in cropped_ronc.shape]):
            warn("Requested crop ammount is greater than the image size.  No cropping will be done.")
            return ronc
        return cropped_ronc

    @staticmethod
    def downSampRoncVec(ronch_vec, binning_factor):
        """
        Downsample the image by taking the mean over nearby values

        Parameters
        ----------
        ronch_vec : ndarray
            Image data
        binning_factor : int
            factor to reduce the size of the image by

        Returns
        -------
        ronc_mat3_mean : ndarray
            Flattened downsampled image
        """
        ccd_pix = int(np.sqrt(ronch_vec.size))
        ronc_mat = ronch_vec.reshape(ccd_pix, ccd_pix)
        ronc_mat2 = ronc_mat.reshape(ccd_pix, ccd_pix / binning_factor, binning_factor)
        ronc_mat2_mean = ronc_mat2.mean(2)  # take the mean along the 3rd dimension
        ronc_mat3 = ronc_mat2_mean.reshape(ccd_pix / binning_factor, binning_factor, -1)
        ronc_mat3_mean = ronc_mat3.mean(1)

        return ronc_mat3_mean.reshape(-1)

    @staticmethod
    def _parse_file_path(image_path):
        """
        Returns a list of all files in the directory given by path

        Parameters
        ---------------
        path : string / unicode
            absolute path to directory containing files

        Returns
        ----------
        file_list : list of strings
            names of all files in directory located at path
        """
        allowed_image_types = ['.ndata1', '.npy']
        if os.path.isdir(image_path):
            # Image path is to a directory
            file_list = list()
            for root, dirs, files in os.walk(image_path):
                for thisfile in files:
                    _, ext = os.path.splitext(thisfile)
                    if ext not in allowed_image_types:
                        continue
                    else:
                        file_list.append(os.path.join(root, thisfile))
        else:
            # Image path is a file
            _, ext = os.path.splitext(image_path)
            if ext in allowed_image_types:
                file_list = [image_path]

        return file_list

    @staticmethod
    def _getimageparms(file_list):
        """
        Returns the image parameters for each file in the `file_list`

        Parameters
        ------------
        file_list : list of zipfile.ZipFile
            List of zipfile objects

        Returns
        -----------
        parm_list : list of dict
            List of image parameters from the files in `file_list`
        """
        parm_list = list()

        for fpath in file_list:
            base, ext = os.path.splitext(fpath)
            if ext == '.ndata1':
                zfile = zipfile.ZipFile(fpath, 'r')
                tmp_path = zfile.extract('metadata.json')
            elif ext == '.npy':
                folder, basename = os.path.split(base)
                same_name_path = base+'.json'
                metapath = os.path.join(folder, 'metadata.json')
                if os.path.exists(same_name_path):
                    tmp_path = same_name_path
                elif os.path.exists(metapath):
                    tmp_path = metapath
            metafile = open(tmp_path, 'r')
            metastring = metafile.read()
            parm_list.append(unnest_parm_dicts(json.loads(metastring)))
            metafile.close()

            if ext == '.ndata1':
                os.remove(tmp_path)

        return parm_list

    def _setupH5(self, image_parms):
        """
        Setup the HDF5 file in which to store the data
        Due to the structure of the ndata format, we can only create the Measurement and Channel groups here

        Parameters
        ----------
        image_parms : dict
            Dictionary of parameters

        Returns
        -------
        h5_main : h5py.Dataset
            HDF5 Dataset that the images will be written into
        h5_mean_spec : h5py.Dataset
            HDF5 Dataset that the mean over all positions will be written
            into
        h5_ronch : h5py.Dataset
            HDF5 Dateset that the mean over all Spectroscopic steps will be
            written into
        """
        root_parms = generate_dummy_main_parms()
        root_parms['data_type'] = 'PtychographyData'

        # Create the hdf5 data Group
        root_grp = MicroDataGroup('/')
        root_grp.attrs = root_parms

        dg_channels = list()
        for meas_parms in image_parms:
            # Create new measurement group for each set of parameters
            meas_grp = MicroDataGroup('Measurement_')
            # Write the parameters as attributes of the group
            for key in meas_parms.keys():
                meas_grp.attrs[key] = meas_parms[key]
            chan_grp = MicroDataGroup('Channel_000')

            meas_grp.addChildren([chan_grp])
            root_grp.addChildren([meas_grp])

        # Write the groups to file
        h5_refs = self.hdf.writeData(root_grp)

        h5_channels = getH5GroupRefs('Channel_000', h5_refs)

        self.hdf.flush()

        return h5_channels

    @staticmethod
    def __no_bin(image, *args, **kwargs):
        """
        Does absolutely nothing to the image.  Exists so that we can have
        a bin function to call whether we actually rebin the image or not.

        Parameters
        ----------
        image : ndarray
            Image
        args:
            Argument list
        kwargs:
            Keyword argument list

        Returns
        -------
        image : ndarray
            The input image
        """
        return image



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  7 21:14:25 2017

@author: Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import h5py
import six
from warnings import warn
import numpy as np
from .hdf_utils import checkIfMain, get_attr, get_data_descriptor, get_formatted_labels, \
    get_dimensionality, get_sort_order, get_unit_values, reshape_to_Ndims
from .io_utils import transformToReal
# from ..viz.jupyter_utils import simple_ndim_visualizer


class PycroDataset(h5py.Dataset):

    def __init__(self, h5_ref, sort_dims=False):
        """
        New data object that extends the h5py.Dataset.

        Parameters
        ----------
        h5_ref : hdf5.Dataset
            The base dataset to be extended
        sort_dims : bool
            Should the dimensions be sorted internally from fastest changing to slowest.

        Methods
        -------
        self.get_current_sorting
        self.toggle_sorting
        self.get_pos_values
        self.get_spec_values
        self.get_n_dim_form
        self.slice


        Attributes
        ----------
        self.h5_spec_vals : h5py.Dataset
            Associated Spectroscopic Values dataset
        self.h5_spec_inds : h5py.Dataset
            Associated Spectroscopic Indices dataset
        self.h5_pos_vals : h5py.Dataset
            Associated Position Values dataset
        self.h5_pos_inds : h5py.Dataset
            Associated Position Indices dataset
        self.pos_dim_labels : list of str
            The labels for the position dimensions.
        self.spec_dim_labels : list of str
            The labels for the spectroscopic dimensions.
        self.n_dim_labels : list of str
            The labels for the n-dimensional dataset.
        self.pos_dim_sizes : list of int
            A list of the sizes of each position dimension.
        self.spec_dim_sizes : list of int
            A list of the sizes of each spectroscopic dimension.
        self.n_dim_sizes : list of int
            A list of the sizes of each dimension.

        Notes
        -----
        The order of all labels and sizes attributes is determined by the current value of `sort_dims`.

        """

        if not checkIfMain(h5_ref):
            raise TypeError('Supply a h5py.Dataset object that is a pycroscopy main dataset')

        super(PycroDataset, self).__init__(h5_ref.id)

        # User accessible properties
        # The required Position and Spectroscopic datasets
        self.h5_spec_vals = self.file[self.attrs['Spectroscopic_Values']]
        self.h5_spec_inds = self.file[self.attrs['Spectroscopic_Indices']]
        self.h5_pos_vals = self.file[self.attrs['Position_Values']]
        self.h5_pos_inds = self.file[self.attrs['Position_Indices']]

        # The dimension labels
        self.__pos_dim_labels = get_attr(self.h5_pos_inds, 'labels')
        self.__spec_dim_labels = get_attr(self.h5_spec_inds, 'labels')

        # Data desciptors
        self.data_descriptor = get_data_descriptor(self)
        self.pos_dim_descriptors = get_formatted_labels(self.h5_pos_inds)
        self.spec_dim_descriptors = get_formatted_labels(self.h5_spec_inds)

        # The size of each dimension
        self.__pos_dim_sizes = np.array(get_dimensionality(np.transpose(self.h5_pos_inds)))
        self.__spec_dim_sizes = np.array(get_dimensionality(np.atleast_2d(self.h5_spec_inds)))

        # Sorted dimension order
        self.__pos_sort_order = get_sort_order(np.transpose(self.h5_pos_inds))
        self.__spec_sort_order = get_sort_order(np.atleast_2d(self.h5_spec_inds))

        # iternal book-keeping / we don't want users to mess with these?
        self.__n_dim_sizes = np.append(self.__pos_dim_sizes, self.__spec_dim_sizes)
        self.__n_dim_labs = np.append(self.__pos_dim_labels, self.__spec_dim_labels)
        self.__n_dim_sort_order = np.append(self.__pos_sort_order, self.__spec_sort_order)
        self.__n_dim_data = None

        # Should the dimensions be sorted from fastest to slowest
        self.__sort_dims = sort_dims

        self.__set_labels_and_sizes()

    def __eq__(self, other):
        if isinstance(other, PycroDataset):
            if isinstance(other, h5py.Dataset):
                warn('Comparing PycroData object with h5py.Dataset')

            return super(PycroDataset, self).__eq__(other)

        return False

    def __repr__(self):
        h5_str = super(PycroDataset, self).__repr__()

        pos_str = ' \n'.join(['{} - size: {}'.format(dim_name, str(dim_size)) for dim_name, dim_size in
                              zip(self.__pos_dim_labels, self.__pos_dim_sizes)])
        spec_str = ' \n'.join(['{} - size: {}'.format(dim_name, str(dim_size)) for dim_name, dim_size in
                               zip(self.__spec_dim_labels, self.__spec_dim_sizes)])

        pycro_str = ' \n'.join(['located at:',
                                self.name,
                                'Data contains:', self.data_descriptor,
                                'Data dimensions and original shape:',
                                'Position Dimensions:',
                                pos_str,
                                'Spectroscopic Dimensions:',
                                spec_str])

        r = '\n'.join([h5_str, pycro_str])

        if six.PY2:
            pycro_str = pycro_str.encode('utf8')

        return '\n'.join([h5_str, pycro_str])

    def __set_labels_and_sizes(self):
        """
        Sets the labels and sizes attributes to the correct values based on
        the value of `self.__sort_dims`

        Returns
        -------
        None

        """
        if self.__sort_dims:
            self.pos_dim_labels = self.__pos_dim_labels[self.__pos_sort_order].tolist()
            self.spec_dim_labels = self.__spec_dim_labels[self.__spec_sort_order].tolist()
            self.pos_dim_sizes = self.__pos_dim_sizes[self.__pos_sort_order].tolist()
            self.spec_dim_sizes = self.__spec_dim_sizes[self.__spec_sort_order].tolist()
            self.n_dim_labels = self.__n_dim_labs[self.__n_dim_sort_order].tolist()
            self.n_dim_sizes = self.__n_dim_sizes[self.__n_dim_sort_order].tolist()

        else:
            self.pos_dim_labels = self.__pos_dim_labels.tolist()
            self.spec_dim_labels = self.__spec_dim_labels.tolist()
            self.pos_dim_sizes = self.__pos_dim_sizes.tolist()
            self.spec_dim_sizes = self.__spec_dim_sizes.tolist()
            self.n_dim_labels = self.__n_dim_labs.tolist()
            self.n_dim_sizes = self.__n_dim_sizes.tolist()

    def get_pos_values(self, dim_name):
        """
        Extract the values for the specified position dimension

        Parameters
        ----------
        dim_name : str
            Name of one of the dimensions in `self.pos_dim_labels`

        Returns
        -------
        dim_values : numpy.ndarray
            Array containing the unit values of the dimension `dim_name`

        """
        return get_unit_values(self.h5_pos_inds, self.h5_pos_vals)[dim_name]

    def get_spec_values(self, dim_name):
        """
        Extract the values for the specified spectroscopic dimension

        Parameters
        ----------
        dim_name : str
            Name of one of the dimensions in `self.spec_dim_labels`

        Returns
        -------
        dim_values : numpy.ndarray
            Array containing the unit values of the dimension `dim_name`

        """
        return get_unit_values(self.h5_spec_inds, self.h5_spec_vals)[dim_name]

    def get_current_sorting(self):
        """
        Prints the current sorting method.

        """
        if self.__sort_dims:
            print('Data dimensions are sorted in order from fastest changing dimension to slowest.')
        else:
            print('Data dimensions are in the order they occur in the file.')

    def toggle_sorting(self):
        """
        Toggles between sorting from the fastest changing dimension to the slowest and sorting based on the
        order of the labels

        """
        if self.__n_dim_data is not None:
            if self.__sort_dims:
                nd_sort = self.__pos_sort_order[::-1] + self.__spec_sort_order[::-1]
            else:
                nd_sort = self.__n_dim_sort_order

            self.__n_dim_data = np.transpose(self.__n_dim_data, nd_sort)

        self.__sort_dims = not self.__sort_dims

        self.__set_labels_and_sizes()

    def get_n_dim_form(self, as_scalar=False):
        """
        Reshapes the dataset to an N-dimensional array

        Returns
        -------
        n_dim_data : numpy.ndarray
            N-dimensional form of the dataset

        """

        n_dim_data, success = reshape_to_Ndims(self, sort_dims=self.__sort_dims)

        if success is not True:
            raise ValueError('Unable to reshape data to N-dimensional form.')

        if as_scalar:
            self.__n_dim_data = transformToReal(n_dim_data)
        else:
            self.__n_dim_data = n_dim_data

        return self.__n_dim_data

    def slice(self, as_scalar=False, slice_dict=dict()):
        """
        Slice the dataset based on an input dictionary of 'str': slice pairs.
        Each string should correspond to a dimension label.  The slices can be
        array-likes or slice objects.

        Parameters
        ----------
        as_scalar : bool
            Should the data be returned as scalar values only.
        slice_dict : dict
            Dictionary of array-likes.

        Returns
        -------
        data_slice : numpy.ndarray
            Slice of the dataset.  Dataset has been reshaped to N-dimensions if `success` is True, only
            by Position dimensions if `success` is 'Positions', or not reshape at all if `success`
            is False.
        success : str or bool
            Informs the user as to how the data_slice has been shaped.

        """
        # Convert the slice dictionary into lists of indices for each dimension
        pos_slice, spec_slice = self.get_pos_spec_slices(slice_dict)

        # Now that the slices are built, we just need to apply them to the data
        # This method is slow and memory intensive but shouldn't fail if multiple lists are given.
        if len(pos_slice) <= len(spec_slice):
            # Fewer final positions that spectra (Most common case)
            data_slice = np.atleast_2d(self[pos_slice, :])[:, spec_slice]
        else:
            data_slice = np.atleast_2d(self[:, spec_slice])[pos_slice, :]

        pos_inds = self.h5_pos_inds[pos_slice, :]
        spec_inds = self.h5_spec_inds[:, spec_slice].reshape([self.h5_spec_inds.shape[0], -1])
        data_slice, success = reshape_to_Ndims(data_slice,
                                               h5_pos=pos_inds,
                                               h5_spec=spec_inds)

        if as_scalar:
            return transformToReal(data_slice), success
        else:
            return data_slice, success

    def get_pos_spec_slices(self, slice_dict):
        """
        Convert the slice dictionary into two lists of indices, one each for the position and spectroscopic
        dimensions.

        Parameters
        ----------
        slice_dict : dict
            Dictionary of array-likes.

        Returns
        -------
        pos_slice : list of uints
            Position indices included in the slice
        spec_slice : list of uints
            Spectroscopic indices included in the slice

        """
        # Create default slices that include the entire dimension
        n_dim_slices = dict()
        n_dim_slices_sizes = dict()
        for dim_lab, dim_size in zip(self.n_dim_labels, self.n_dim_sizes):
            n_dim_slices[dim_lab] = list(range(dim_size))
            n_dim_slices_sizes[dim_lab] = len(n_dim_slices[dim_lab])
        # Loop over all the keyword arguments and create slices for each.
        for key, val in slice_dict.items():
            # Make sure the dimension is valid
            if key not in self.__n_dim_labs:
                raise KeyError('Cannot slice on dimension {}.  '
                               'Valid dimensions are {}.'.format(key, self.__n_dim_labs.tolist()))

            # Check the value and convert to a slice object if possible.
            # Use a list if not.
            if isinstance(val, slice):
                val = n_dim_slices[key][val]
            elif isinstance(val, list):
                pass
            elif isinstance(val, np.ndarray):
                val = val.flatten().tolist()
            elif isinstance(val, tuple):
                val = list(val)
            else:
                raise TypeError('The slices must be array-likes or slice objects.')

            n_dim_slices[key] = val

            n_dim_slices_sizes[key] = len(val)

        # Build the list of position slice indices
        for pos_ind, pos_lab in enumerate(self.__pos_dim_labels):
            n_dim_slices[pos_lab] = np.isin(self.h5_pos_inds[:, pos_ind], n_dim_slices[pos_lab])
            if pos_ind == 0:
                pos_slice = n_dim_slices[pos_lab]
            else:
                pos_slice = np.logical_and(pos_slice, n_dim_slices[pos_lab])
        pos_slice = np.argwhere(pos_slice)

        # Do the same for the spectroscopic slice
        for spec_ind, spec_lab in enumerate(self.__spec_dim_labels):
            n_dim_slices[spec_lab] = np.isin(self.h5_spec_inds[spec_ind], n_dim_slices[spec_lab])
            if spec_ind == 0:
                spec_slice = n_dim_slices[spec_lab]
            else:
                spec_slice = np.logical_and(spec_slice, n_dim_slices[spec_lab])
        spec_slice = np.argwhere(spec_slice)

        return pos_slice, spec_slice

    # def visualize(self, slice_dict=None, **kwargs):
    #     """
    #     Interactive visualization of this dataset. Only available on jupyter notebooks
    #
    #     Parameters
    #     ----------
    #     slice_dict : dictionary, optional
    #         Slicing instructions
    #     """
    #     # TODO: Robust implementation that allows slicing
    #     if len(self.pos_dim_labels + self.spec_dim_labels) > 4:
    #         raise NotImplementedError('Unable to support visualization of more than 4 dimensions. Try slicing')
    #     data_mat = self.get_n_dim_form()
    #     pos_dim_names = self.pos_dim_labels[::-1]
    #     spec_dim_names = self.spec_dim_labels
    #     pos_dim_units_old = get_attr(self.h5_pos_inds, 'units')
    #     spec_dim_units_old = get_attr(self.h5_spec_inds, 'units')
    #     pos_ref_vals = get_unit_values(self.h5_pos_inds, self.h5_pos_vals, is_spec=False)
    #     spec_ref_vals = get_unit_values(self.h5_spec_inds, self.h5_spec_vals, is_spec=True)
    #
    #     simple_ndim_visualizer(data_mat, pos_dim_names, pos_dim_units_old, spec_dim_names, spec_dim_units_old,
    #                            pos_ref_vals=pos_ref_vals, spec_ref_vals=spec_ref_vals, **kwargs)
